"""Element library
"""
import inspect
from functools import partial
from abc import abstractmethod

import numpy as np
from numpy.polynomial import Polynomial
from scipy.integrate import quad
import scipy.integrate
# import tensorflow as tf

import anon
from emme.matrices import Structural_Matrix, Structural_Vector
try:
    import anon.atom as anp
except:
    anp = np
# dtype = 'float32'
# int_dtype='int32'
# tf.keras.backend.set_floatx(dtype)

class IntForce:
    def __init__(self, elem:object, number:int, nature:str=None):
        self.number:int = number
        self.nature:str = nature
        self.elem = elem
        self.rel = False
        self.redundant = False
        self.plastic_event = None

    @property
    def tag(self)->str:
        return self.elem.tag + '_' + str(self.number)

    def __str__(self):
        return self.tag

    def __repr__(self):
        return self.tag


class BasicLink():
    """Class implementing general geometric element methods"""

    def __init__(self, ndf, ndm, nodes):
        self.nodes = nodes
        self.ndf: int = ndf
        self.ndm: int = ndm
        self.nen = len(nodes)

    @property
    def loc(self)->anp.ndarray:
        return anp.array([node.xyz for node in self.nodes])

    @property
    def dofs(self):
        """

        This function is very slow.
        """
        eid = np.array([],dtype=int)
        for node in self.nodes:
            eid = np.append(eid, node.dofs[0:self.ndf])
        #untested alternative:
        #return np.array([node.dofs[0:self.ndf] for node in self.nodes])
        return eid

    @property
    def L(self):
        xyzi = self.nodes[0].xyz
        xyzj = self.nodes[1].xyz
        L = np.linalg.norm(xyzi-xyzj)
        return L

    @property
    def L0(self):
        xyzi = self.nodes[0].xyz0
        xyzj = self.nodes[1].xyz0
        L = np.linalg.norm(xyzi-xyzj)
        return L

    @property
    def Li(self):
        n1 = self.nodes[0]
        n2 = self.nodes[1]
        xyzi = np.array([n1.xi, n1.yi, n1.zi])
        xyzj = np.array([n2.xi, n2.yi, n2.zi])
        L = np.linalg.norm(xyzi-xyzj)
        return L


    @property
    def Dx(self)->float:
        return self.nodes[1].x - self.nodes[0].x

    @property
    def Dy(self)->float:
        return self.nodes[1].y - self.nodes[0].y

    @property
    def Dz(self)->float:
        return self.nodes[1].z - self.nodes[0].z

    @property
    def sn(self)->float:
        """directional sine"""
        L = self.L
        sn = (self.nodes[1].y - self.nodes[0].y)/L
        return sn

    @property
    def cs(self):
        """directional cosine"""
        L = self.L
        cs = (self.nodes[1].x - self.nodes[0].x)/L
        return cs

    @property
    def cz(self):
        L = self.L
        cz = (self.nodes[1].z - self.nodes[0].z)/L
        return cz

    def Rx_matrix(self):
        """Rotation about x

        """
        cs = self.cs
        sn = self.sn
        Rx = np.array([
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0,  cs, -sn, 0.0, 0.0, 0.0],
            [0.0,  sn,  cs, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0,  cs, -sn],
            [0.0, 0.0, 0.0, 0.0,  sn,  cs],
        ])

        return 0

    def Ry_matrix(self):
        "Rotation about z"
        cs = self.cs
        sn = self.sn
        Ry = np.array([
            [ cs, 0.0,  sn],
            [0.0, 1.0, 0.0],
            [-sn, 0.0,  cs],
        ])

        return 0

    def Rz_matrix(self)->np.ndarray:
        "Rotation about z"
        cs = self.cs
        sn = self.sn
        Rz = np.array([
            [ cs, -sn, 0.0, 0.0, 0.0, 0.0],
            [ sn,  cs, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0,  cs, -sn, 0.0],
            [0.0, 0.0, 0.0,  sn,  cs, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 1.0]])

        return Rz


class Element(BasicLink):
    """Element parent class"""

    def __init__(self,  ndf, ndm, force_dict=None, nodes=None, elem=None, resp=None, proto=None, tag=None):
        super().__init__(ndf, ndm, nodes)
        self._resp = resp
        self.elem = elem
        self.history = {}
        self.current = {}
        if force_dict is None:
            force_dict = {str(i+1): 0 for i in range(ndf)}
        if tag is not None:
            self.tag = tag

        nq = len(force_dict)
        self.rel = {str(i): False for i in range(1, nq+1)}
        self.red = {str(i): False for i in range(1, nq+1)}
        self.q0  = {str(i): 0.  for i in range(1, nq+1)}
        self.v0  = {str(i): 0.  for i in range(1, nq+1)}
        self.e0  = {str(i): 0.  for i in range(1, nq+1)}
        self.Qp  = {'+': {str(i): 0.  for i in range(1, nq+1)}, '-':{str(i): 0.  for i in range(1, nq+1)}}

        self.basic_forces = np.array([IntForce(self, i, nature=str(i)) for i in range(1, nq+1)])
        self.basic_deformations = self.basic_forces

    @property
    def nn(self): return len(self.nodes)

    @property
    def resp(self):
        return self._resp

    @resp.setter
    def resp(self, value):
        self._resp = value

    @property
    def force_keys(self):
        return [self.tag+'_'+key for key in self.rel]

    def compose(self, **model_params):
        """
        created 2021-03-31
        """
        if self.elem is None:
            # create a default linear element
            ke = anp.array(self.ke_matrix())
            def f(x,y=None,state={},params={},**kwds):
                return None,ke@x,state
            stiff = lambda x,y,state={},params={},**kwds: ke
            return anon.dual.wrap(f,dim=(self.ndf,1),jacx=stiff)

        else:
            elem = self.elem
            #parameters = inspect.signature(elem).parameters
            #for name,param in model_params.items():
            #    parameters[name].replace(default=param)

            return elem



    def v0_vector(self):
        return np.array([0]*self.ndf*self.nn)

    def pw_vector(self):
        if all(self.w.values())==0.0:
            return np.array([0.0]*self.ndf*self.nn)

class PolyRod(Element):
    nv  = 1
    nn  = 2
    ndm = 1
    ndf = 1 # number of dofs at each node
    force_dict = {'1':0}

    def __init__(self,tag, nodes, E, A):
        super().__init__(self.ndf, self.ndm, self.force_dict, nodes)
        self.tag:str = tag
        self.E:float = E
        """Young's modulus of elasticity"""
        self.A:float = A
        """cross-sectional areal"""
        self.q = {'1':0}
        self.v = {'1':0}
        self.w = {'1':0.0}

        if isinstance(self.E,float):
            self.E = Polynomial([self.E])

        if isinstance(self.A,float):
            self.A = Polynomial([self.A])
    
    def N(self):
        L = self.L
        N1 = Polynomial([1,-1/L])
        N2 = Polynomial([0, 1/L])
        return np.array([[N1],[N2]])
    
    def B(self):
        N = self.N()
        return np.array([[Polynomial.deriv(Ni,1) for Ni in row] for row in N])

    def k_matrix(self):
        E = self.E
        A = self.A

        L = self.L
        B = self.B()
        k = Structural_Matrix([
            [quad(E*A*(B@B.T)[i,j],0,L)[0] for j in range(2)] for i in range(2)
        ])

        # Metadata
        k.tag = 'k'
        k.row_data = k.column_data = ['u_'+str(int(dof)) for dof in self.dofs]
        return k

    def ke_matrix(self):
        return self.k_matrix()

    def pw_vector(self):
        L = self.L
        pw = self.w['1']
        N = self.N()
        if isinstance(pw,np.polynomial.Polynomial) or isinstance(pw,float):
            # p1 = -quad(N[0]*pw,0,L)[0]
            # p2 = -quad(N[1]*pw,0,L)[0]
            p = np.array([[-quad(Ni*pw,0,L)[0] for Ni in row] for row in N])
        else:
            print('Unsupported load vector pw')

        return p

    ## Post Processing
    def localize(self,U_vector):
        dofs = [int(dof) for dof in self.dofs]
        return np.array([U_vector.get(str(dof)) for dof in dofs])
    
    def displx(self,U_vector):
        u = self.localize(U_vector)
        N = self.N()
        return N[0,0]*u[0] + N[1,0]*u[1]
        
    def strainx(self,U_vector):
        dofs = [int(dof) for dof in self.dofs]
        u = np.array([U_vector.get(str(dof)) for dof in dofs])
        B = self.B()
        return B[0,0]*u[0] + B[1,0]*u[1]

    def iforcex(self,U_vector):
        """Resisting force vector"""
        dofs = [int(dof) for dof in self.dofs]
        u = np.array([U_vector.get(str(dof)) for dof in dofs])
        B = self.B()
        P = self.E*self.A*(B[0,0]*u[0] + B[1,0]*u[1])
        return P

class MeshCell(Element):
    def __init__(self, tag, nn, ndf, ndm, nodes):
        node_map = {}
        super().__init__(ndf, ndm, nodes=nodes, tag=tag)
    
    @property
    def dofs(self):
        """"""
        return np.asarray([node.dofs[0:self.ndf] for node in self.nodes])


    def ke_matrix(self, *args, **kwds):
        return anp.eye(self.ndm*self.nn)

class Truss(Element):
    nv = 1
    nn = 2
    nq = 1
    ndm = 2
    ndf = 2 #: number of dofs at each node
    force_dict = {'1':0}
    Qpl = np.zeros((2,nq))

    def __init__(self, tag, iNode, jNode, E=None, A=None, geo='lin',properties=None,**kwds):
        super().__init__(self.ndf, self.ndm, self.force_dict, [iNode, jNode],**kwds)
        if isinstance(properties,dict):
            E, A = properties['E'], properties['A']
        self.type = '2D truss'
        self.tag = tag
        self.E = E
        self.A = A
        self.geo=geo
        self.Q = np.array([0.0])

        self.q = {'1':0}
        self.v = {'1':0}
        
        self.w = {'1':0.0}

    def __repr__(self):
        return 'truss-{}'.format(self.tag)
    
    def N(self):
        L = self.L
        N1 = Polynomial([1,-1/L])
        N2 = Polynomial([0, 1/L])
        return np.array([N1,N2])
    
    def B(self):
        L = self.L
        B1 = self.N()[0].deriv(1)
        B2 = self.N()[1].deriv(1)
        return np.array([B1,B2])

    def v0_vector(self):
        EA = self.E*self.A
        L = self.L
        e0 = self.e0
        q0 = self.q0
        w = self.w
        v0 =  np.array([e0['1']*L])
        v0 += [q0['1']*L/EA]
        v0 += [w['1']*L*L/(2*EA)]
        return v0
        
    def q0_vector(self):
        EA = self.E*self.A
        e0 = self.e0['1']
        q0 = self.q0['1']
        q0 = q0 - EA*e0
        return [q0]

    def pw_vector(self):
        L = self.L
        pw = self.w['1']
        N = self.N()
        if isinstance(pw,np.polynomial.Polynomial) or isinstance(pw,float):
            p1 = -quad(N[0]*pw,0,L)[0]
            p2 = -quad(N[1]*pw,0,L)[0]
        else:
            print('Unsupported load vector pw')

        return np.array([p1,0,p2,0])

    def ke_matrix(self):
        ag = self.ag()
        k = self.k_matrix()
        return ag.T@(k*ag)

    def kg_matrix(self, N):
        """return element local stiffness matrix"""
        E = self.E
        A = self.A
        L = self.L
        k = Structural_Matrix([N/L])
        # Metadata
        k.tag = 'kg'
        k.row_data = ['q_'+ key for key in self.q0.keys()]
        k.column_data = ['v_' + key for key in self.e0.keys()]
        k.c_cidx = k.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]
        return k

    def k_matrix(self): 
        """return element local stiffness matrix"""

        E = self.E
        A = self.A
        L = self.L
        k = Structural_Matrix([E*A/L])
        # Metadata
        k.tag = 'k'
        k.row_data = ['q_'+ key for key in self.q0.keys()]
        k.column_data = ['v_' + key for key in self.e0.keys()]
        k.c_cidx = k.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]
        return k

    def bg_matrix(self, **kwds):
        """return element static matrix, $\\mathbf{b}_g$"""

        cs = self.cs
        sn = self.sn
        bg = np.array([[-cs],
                       [-sn],
                       [ cs],
                       [ sn]])
        return bg

    def f_matrix(self, Roption=True):
        """return element flexibility matrix, $\\mathbf{f}$"""
        
        A = self.A
        E = self.E
        L = self.L
        f = Structural_Matrix([L/(E*A)])

        if Roption:
            pass

        # Metadata
        f.tag = 'f'
        f.column_data = ['q_' + key for key in self.q0.keys()]
        f.row_data = ['v_'+ key for key in self.e0.keys()]
        f.c_cidx = f.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]

        return f
    
    def ag(self): 
        cs = self.cs
        sn = self.sn
        ag = np.array([[-cs,-sn , cs, sn],])
        return ag
    

    def GLstrain(self):
        Li = self.Li
        L0 = self.L0
        E_GL = (Li**2 - L0**2) / (2*L0**2)
        return E_GL
    
    def iGLstrain(self):
        """incremental Green-Lagrange strain"""
        L  = self.L
        Li = self.Li
        E_GL = (L**2 - Li**2) / (2*Li**2)
        return E_GL

class TaperedTruss(Truss):
    def k_matrix(self):
        if isinstance(self.E,float):
            E = Polynomial([self.E])
        else:
            E = self.E
        if isinstance(self.A,float):
            A = Polynomial([self.A])
        else:
            A = self.A
        A = self.A 
        L = self.L
        B = self.B()
        # ke = np.zeros((self.ndf,self.ndf))
        # for i in range(self.ndf):
        #     for j in range(self.ndf):
        #         f = lambda x: B[i](x)*E(x)*A(x)*B[j](x)
        #         ke[i,j] = quad(f,0,L)[0]

        f = lambda x: B[0](x)*E(x)*A(x)*B[0](x)
        k = Structural_Matrix([quad(f,0,L)[0]])
        
        # Metadata
        k.tag = 'k'
        k.row_data = ['q_'+ key for key in self.q0.keys()]
        k.column_data = ['v_' + key for key in self.e0.keys()]
        k.c_cidx = k.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]
        return k
    
    def q0_vector(self):
        # EA = self.E*self.A()
        # e0 = self.e0['1']
        # q0 = self.q0['1']
        # q0 = q0 - EA*e0
        return [0]
    
    def v0_vector(self):
        return [0]

class Truss3D(Element):
    ndf = 3
    ndm = 3
    force_dict = {'1':0}
    def __init__(self, tag, iNode, jNode, A, E):
        super().__init__(self.ndf, self.ndm, self.force_dict, [iNode,jNode])
        self.type = '2D truss'
        self.tag = tag
        self.A = A
        self.E = E

        self.q = {'1':0}
        self.v = {'1':0}
    
    def __repr__(self):
        return 'tr-{}'.format(self.tag)
           
    def bg_matrix(self): 
        """return element static matrix, bg - pp. 57"""
        cs = self.cs
        sn = self.sn
        cz = self.cz
        bg = np.array([[-cs],
                       [-sn],
                       [-cz],
                       [ cs],
                       [ sn],
                       [ cz]])
        return bg


class Beam(Element): 
    """linear 2D Euler-Bernouli frame element"""

    nv = 3
    nn = 2
    nq = 3
    ndm = 2
    ndf = 3
    force_dict = {'1':0, '2':0, '3': 0}
    Qpl = np.zeros((2,nq))

    def __init__(self, tag, iNode, jNode, E=None, A=None, I=None, properties=None, **kwds):
        #print(kwds)
        super().__init__(self.ndf, self.ndm, self.force_dict, [iNode,jNode], **kwds)
        if isinstance(properties,dict):
            E, A, I = properties['E'], properties['A'], properties['I']
        self.type = '2D beam'
        self.tag:str = tag
        self.E:float = E
        """Young's modulus of elasticity"""
        self.A:float = A
        """Cross-sectional area"""
        self.I:float = I
        """Cross-sectional moment of inertia"""
        self.q = {'1':0, '2':0, '3': 0}     # basic element force
        self.v = {'1':0, '2':0, '3': 0}
        self.w = {'x':0, 'y':0}             #':  "uniform element loads in x and y",

    def __repr__(self):
        return 'el-{}'.format(self.tag)

    def Elastic_curve(self, x, end_rotations, scale:float=10, global_coord=False):
        n = len(x)
        L = self.L
        N1 = 1-3*(x/L)**2+2*(x/L)**3
        N2 = L*(x/L-2*(x/L)**2+(x/L)**3)
        N3 = 3*(x/L)**2-2*(x/L)**3
        N4 = L*((x/L)**3-(x/L)**2)
        vi = end_rotations[0]
        vj = end_rotations[1]
        y = np.array(vi*N2+vj*N4)*scale
        xy = np.concatenate(([x],[y]))
        if global_coord:
            x0 = self.nodes[0].x
            y0 = self.nodes[0].y
            Rz = self.Rz_matrix()[0:2,0:2]
            xy = Rz@xy + [[x0]*n,[y0]*n]
        return xy

    @property
    def enq(self): 
        """element number of forces, considering deprecation"""
        # output like [1, 1, 1]
        return [1 for x in self.q]

    def ag(self):
        cs = self.cs
        sn = self.sn
        L = self.L
        ag = np.array([[ -cs ,  -sn , 0,  cs ,   sn , 0],
                       [-sn/L,  cs/L, 1, sn/L, -cs/L, 0],
                       [-sn/L,  cs/L, 0, sn/L, -cs/L, 1]])
        
        if self.dofs[0] == self.dofs[3] or self.dofs[1] == self.dofs[4]:
            ag[0,:] = [0.0]*ag.shape[1]
        return ag

    def ah(self):
        MR = [1 if x else 0 for x in self.rel.values()]
        ah = np.array([[1-MR[0],          0          ,             0        ],
                       [   0   ,       1-MR[1]       ,  -0.5*(1-MR[2])*MR[1]],
                       [   0   , -0.5*(1-MR[1])*MR[2],          1-MR[2]     ]])
        return ah

    def k_matrix(self,*args,**kwds): 
        """return element local stiffness Matrix"""
        E = self.E
        A = self.A
        I = self.I
        L = self.L
        EI = E*I
        ah = self.ah()
        k = np.array([[E*A/L,    0   ,   0   ],
                      [  0  , 4*EI/L , 2*EI/L],
                      [  0  , 2*EI/L , 4*EI/L]])
        k = ah.T @ k @ ah

        # Assemble matrix metadata
        k = Structural_Matrix(k)
        k.tag = 'k'
        k.row_data = ['q_'+ key for key in self.q0.keys()]
        k.column_data = ['v_' + key for key in self.v0.keys()]
        k.c_cidx = k.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]
        return k

    def f_matrix(self, Roption=False):
        """Flexibility matrix of an element.

        """
        EA = self.E*self.A
        EI = self.E*self.I
        L = self.L
        f = Structural_Matrix([
            [L/EA,     0    ,      0   ],
            [  0 ,  L/(3*EI), -L/(6*EI)],
            [  0 , -L/(6*EI),  L/(3*EI)]])
        ide = set(range(3))

        if Roption:  
            if self.rel["2"]:
                f[:,1] = [0.0]*f.shape[0]
                f[1,:] = [0.0]*f.shape[1]

            if self.rel["3"]:
                f[:,2] = [0.0]*f.shape[0]
                f[2,:] = [0.0]*f.shape[1]

        # Define matrix metadata
        f.tag = 'f'
        f.column_data = ['q_'+ key for key in self.q0.keys()]
        f.row_data = ['v_' + key for key in self.v0.keys()]
        f.c_cidx = f.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]
        return f

    def ke_matrix(self,*args,**kwds):
        """return element global stiffness Matrix"""

        k  = self.k_matrix()
        ag = self.ag()

        ke = ag.T @ k @ ag
        ke = Structural_Matrix(ke)
        ke.row_data = ke.column_data = ['u_'+str(int(dof)) for dof in self.dofs]
        return ke

    def bg_matrix(self, Roption=False):
        """return element global static matrix, bg"""

        cs = self.cs
        sn = self.sn
        L  = self.L
        #                x      ri      rj   Global
        bg = np.array([[-cs, -sn/L,  -sn/L],  # x
                       [-sn,  cs/L,   cs/L],  # y
                       [0.0,   1.0,    0.0],  # rz
                       [ cs,  sn/L,   sn/L],
                       [ sn, -cs/L,  -cs/L],
                       [0.0,   0.0,    1.0]])
        if Roption:
            if self.rel['2']:
                bg[:,1] = [0.0]*bg.shape[0]

            if self.rel['3']:
                bg[:,2] = [0.0]*bg.shape[0]

        if self.dofs[0] == self.dofs[3] or self.dofs[1] == self.dofs[4]:
            bg[:,0] = [0.0]*bg.shape[0]
            # bg = np.delete(bg, 0, axis=1)

        bg = Structural_Matrix(bg)
        bg.tag = 'b'
        bg.column_data = ['x', 'ri', 'rj']
        bg.row_data = ['x', 'y', 'rz']
        bg.c_cidx = bg.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]


        return bg
    
    def m_matrix(self):
        m = np.array([
            [140, 0.0, 0.0,  70.0, 0.0, 0.0],
            [0.0, ],
            [0.0, ],
            [70., ],
            [0.0, ],
            [0.0, ],
        ])
        return m

    def v0_vector(self):
        EA = self.E*self.A
        EI = self.E*self.I
        L = self.L
        e0 = self.e0
        w = self.w
        v0 =  np.array([e0['1']*L, -e0['2']*L/2, e0['2']*L/2])
        v0 += np.array([w['x']*L*L/(2*EA), w['y']*L**3/(24*EI), -w['y']*L**3/(24*EI)])
        v0 = Structural_Matrix(v0)
        v0.tag = 'v'
        v0.column_data = ['v_0']
        v0.row_data = ['axial', 'i-rotation', 'j-rotation']
        v0.c_cidx = False
        v0.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]
        return v0 

    def q0_vector(self):
        L = self.L
        E = self.E 
        A = self.A 
        I = self.I
        e0= self.e0 
        w = self.w
        q0 =  np.array([-e0['1']*E*A, +e0['2']*E*I, -e0['2']*E*I])
        if self.rel['2']:
            q0[1] *= 0
            q0[1] *= 1.5 
        if self.rel['3']:
            q0[2] *= 0
            q0[2] *= 1.5 

        q0[1] += -w['y']*L**2/12*((not self.rel['2']) and (not self.rel['3'])) - w['y']*L**2/8*(self.rel['3'])
        q0[2] += +w['y']*L**2/12*((not self.rel['2']) and (not self.rel['3'])) + w['y']*L**2/8*(self.rel['2'])

        # Metadata
        q0 = Structural_Matrix(q0)
        q0.tag = 'q'
        q0.column_data = ['q_0']
        q0.row_data = ['axial', 'M_i', 'M_j']
        q0.c_cidx = False
        q0.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]

        return q0 

    def κ(self, k, state=None, form=None):
        """Defines element curvature and calculates the resulting end deformations.

        """
        if form ==None: form = 'uniform'
        if state==None: state = -1
        L = self.L
        table = {
            'uniform': [-0.5*k*L, 0.5*k*L],
            'ilinear': [-k/3*L, k/6*L],
            'jlinear': [-k/6*L, k/2*L],
            'parabolic': [-k/3*L, k/3*L],
        }
        self.nodes[0].model.states[state]['v0'][self.tag]['2'] = table[form][0]
        self.nodes[0].model.states[state]['v0'][self.tag]["3"]= table[form][1]
        return


class Beam_Column2D_FL(Beam):
    axial_force = 0.0 


    def k_matrix_exact(self, axial_force=None):
        if axial_force is None: P = self.axial_force
        else: P = axial_force
        L = self.L
        EI = self.E*self.I
        k = np.sqrt(P/EI)
        kL = k*L
        d = 2*(1-np.cos(kL))-(kL*np.sin(kL)) 
        ka = self.E*self.A / L
        kfv = (EI/(L**3))*(kL**3)*np.sin(kL)/d
        kmv = (EI/(L**2))*((kL**2)*(1-np.cos(kL))/d)
        kft = kmv
        kmt =(EI/L)*((kL*(np.sin(kL)-kL*np.cos(kL)))/d)
        kmth = (EI/L)*((kL*(kL-np.sin(kL)))/d)

        eltk = np.array([ 
            [  ka,  0.0,  -0.0, -ka,   0.0,   0.0],
            [ 0.0,  kfv,   kft, -0.0, -kfv,   kft],
            [ 0.0,  kmv,   kmt, -0.0, -kmv,  kmth],
            [ -ka,  0.0,  -0.0,  ka,   0.0,   0.0],
            [ 0.0, -kfv,  -kft, -0.0,  kfv,  -kft],
            [ 0.0,  kft,  kmth, -0.0, -kft,   kmt]])

        # Assemble matrix metadata
        k = Structural_Matrix(eltk)
        k.tag = 'k'
        k.row_data = ['q_'+ key for key in self.q0.keys()]*2
        k.column_data = ['v_' + key for key in self.v0.keys()]*2
        k.c_cidx = k.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]
        return k

    def k_matrix(self, axial_force=None):
        """Evaluates the 2D beam-column stiffness matrix using a 
        power series expansion
        
        """
        if axial_force is None: P = self.axial_force
        else: P = axial_force
        ah = self.ah()
        L = self.L
        EI = self.E*self.I
        k = np.sqrt(P/EI)
        kL = k*L
        ka = self.E*self.A / L

        kmt = 4*EI/L - 2*EI*L/15*k**2 - 11*EI/6300*L**3*k**4
        kmth = 2*EI/L+EI*L/30*k**2 + 13*EI*L**3*k**4/12600
        if P !=0.0:
            d = 2*(1-np.cos(kL))-(kL*np.sin(kL)) 
            kmt =(EI/L)*((kL*(np.sin(kL)-kL*np.cos(kL)))/d)
            kmth = (EI/L)*((kL*(kL-np.sin(kL)))/d)

        k = np.array([ 
            [  ka,   0.0,  0.0],
            [ 0.0,   kmt, kmth],
            [ 0.0,  kmth,  kmt],])

        k = ah.T @ k @ ah

        # Matrix metadata
        k = Structural_Matrix(k)
        k.tag = 'k'
        k.row_data = ['q_'+ key for key in self.q0.keys()]*2
        k.column_data = ['v_' + key for key in self.v0.keys()]*2
        k.c_cidx = k.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]
        return k

    # def ke_matrix(self, axial_force=None):
    #     T = self.Rz_matrix()
    #     k = self.k_matrix(axial_force)
    #     ke = T.T @ k @ T

    #     return ke

class Beam_Column2D(Beam):
    axial_force = 0.0 

    def k_matrix_exact(self, axial_force=None):
        if axial_force is None: P = self.axial_force
        else: P = axial_force
        L = self.L
        EI = self.E*self.I
        k = np.sqrt(abs(P/EI))
        kL = k*L
        d = 2*(1-np.cos(kL))-(kL*np.sin(kL)) 
        ka = self.E*self.A / L
        kfv = (EI/(L**3))*(kL**3)*np.sin(kL)/d
        kmv = (EI/(L**2))*((kL**2)*(1-np.cos(kL))/d)
        kft = kmv
        kmt =(EI/L)*((kL*(np.sin(kL)-kL*np.cos(kL)))/d)
        kmth = (EI/L)*((kL*(kL-np.sin(kL)))/d)

        eltk = np.array([ 
            [  ka,  0.0,   0.0, -ka,   0.0,   0.0],
            [ 0.0,  kfv,   kft,  0.0, -kfv,   kft],
            [ 0.0,  kmv,   kmt,  0.0, -kmv,  kmth],
            [ -ka,  0.0,   0.0,  ka,   0.0,   0.0],
            [ 0.0, -kfv,  -kft,  0.0,  kfv,  -kft],
            [ 0.0,  kft,  kmth,  0.0, -kft,   kmt]])

        # Assemble matrix metadata
        k = Structural_Matrix(eltk)
        k.tag = 'k'
        k.row_data = ['q_'+ key for key in self.q0.keys()]*2
        k.column_data = ['v_' + key for key in self.v0.keys()]*2
        k.c_cidx = k.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]
        return k

    def k_matrix(self, axial_force=None):
        """Evaluates the 2D beam-column stiffness matrix using a 
        power series expansion
        
        """
        if axial_force is None: P = self.axial_force
        else: P = axial_force
        L = self.L
        EI = self.E*self.I
        k = np.sqrt(P/EI)
        kL = k*L
        ka = self.E*self.A / L
        kfv = 12*EI/(L**3)-6*EI/(5*L)*k**2-EI*L/700*k**4
        kmv = 6*EI/L**2 - EI/10*k**2 - EI*L**2/1400*k**4
        kft = kmv
        kmt = 4*EI/L-2*EI*L/15*k**2 - 11*EI/6300*L**3*k**4
        kmth = 2*EI/L+EI*L/30*k**2 + 13*EI*L**3*k**4/12600

        eltk = np.array([ 
            [  ka,  0.0,  -0.0, -ka,   0.0,   0.0],
            [ 0.0,  kfv,   kft, -0.0, -kfv,   kft],
            [ 0.0,  kmv,   kmt, -0.0, -kmv,  kmth],
            [ -ka,  0.0,  -0.0,  ka,   0.0,   0.0],
            [ 0.0, -kfv,  -kft, -0.0,  kfv,  -kft],
            [ 0.0,  kft,  kmth, -0.0, -kft,   kmt]])

        # hinges
        # cols = rows = [int(key)-1 for key in self.rel.keys() if self.rel[key]]
        # f = np.linalg.inv(eltk)

        # for row in rows:
        #     f[row, :] = [0.0]*6
        # for col in cols:
        #     f[:, col] = [0.0]*6
        # k = np.linalg.inv(f)



        # Matrix metadata
        k = Structural_Matrix(eltk)
        k.tag = 'k'
        k.row_data = ['q_'+ key for key in self.q0.keys()]*2
        k.column_data = ['v_' + key for key in self.v0.keys()]*2
        k.c_cidx = k.c_ridx = [int(key)-1 for key in self.rel.keys() if self.rel[key]]
        return k

    def ke_matrix(self, axial_force=None):
        T = self.Rz_matrix()
        k = self.k_matrix(axial_force)
        ke = T.T @ k @ T
        return ke

class Beam3d(Element):
    nv = 6
    nn = 2
    ndm = 3
    ndf = 6
    force_dict = {'1':0, '2':0, '3': 0}
    
    def __init__(self, tag, iNode, jNode, E, A, Iy, Iz, Gs, Kv):
        """
        3D elastic Bernoulli beam element.

        Parameters
        ----------
        E: float
            Young's modulus
        G: float
            Shear modulus
        A: float
            Cross section area
        Iy: float
            Moment of inertia, local y-axis
        Iz: float
            Moment of inertia, local z-axis
        Kv: float
            Saint-Venant's torsion constant

        """
        super().__init__(self.ndf, self.ndm, self.force_dict, [iNode,jNode])
        self.type = '3D beam'
        self.tag = tag
        self.E:float  =  E
        """Young's modulus of elasticity"""
        self.A  =  A
        self.Gs = Gs
        self.Iy = Iy
        self.Iz = Iz
        self.Kv = Kv
        # self.q = {'1':0, '2':0, '3': 0}     # basic element force
        # self.v = {'1':0, '2':0, '3': 0}

        self.w = {'x':0, 'y':0, 'z': 0}       #':  "uniform element loads in x and y",

         
    def __repr__(self):
        return 'el-{}'.format(self.tag)

    def k_matrix(self):
        """
        Calculate the stiffness matrix for a 3D elastic Bernoulli
        beam element.

        Parameters
        ----------
        E: float
            Young's modulus
        G: float
            Shear modulus
        A: float
            Cross section area
        Iy: float
            Moment of inertia, local y-axis
        Iz: float
            Moment of inertia, local z-axis
        Kv: float
            Saint-Venant's torsion constant

        Returns:
            Kle                      local beam stiffness matrix (12 x 12)
        """
        
        L,E,Gs,A,Iy,Iz,Kv = self.L, self.E, self.Gs, self.A, self.Iy, self.Iz, self.Kv

        a = E*A/L
        b = 12*E*Iz/L**3
        c = 6*E*Iz/L**2
        d = 12*E*Iy/L**3
        e = 6*E*Iy/L**2
        f = Gs*Kv/L
        g = 2*E*Iy/L
        h = 2*E*Iz/L

        Kle = np.mat([
            [ a,  0,  0,  0,  0,   0,  -a,  0,  0,  0,  0,   0  ],
            [ 0,  b,  0,  0,  0,   c,   0, -b,  0,  0,  0,   c  ],
            [ 0,  0,  d,  0, -e,   0,   0,  0, -d,  0, -e,   0  ],
            [ 0,  0,  0,  f,  0,   0,   0,  0,  0, -f,  0,   0  ],
            [ 0,  0, -e,  0, 2*g,  0,   0,  0,  e,  0,  g,   0  ],
            [ 0,  c,  0,  0,  0,   2*h, 0, -c,  0,  0,  0,   h  ],
            [-a,  0,  0,  0,  0,   0,   a,  0,  0,  0,  0,   0  ],
            [ 0, -b,  0,  0,  0,  -c,   0,  b,  0,  0,  0,  -c  ],
            [ 0,  0, -d,  0,  e,   0,   0,  0,  d,  0,  e,   0  ],
            [ 0,  0,  0, -f,  0,   0,   0,  0,  0,  f,  0,   0  ],
            [ 0,  0, -e,  0,  g,   0,   0,  0,  e,  0, 2*g,  0  ],
            [ 0,  c,  0,  0,  0,   h,   0, -c,  0,  0,  0,  2*h ]])

        # fle = L/2*np.mat([qx, qy, qz, qw, -qz*L/6, qy*L/6, qx, qy, qz, qw, qz*L/6, -qy*L/6]).T

        # n2 = np.array([0.,0.,0.])
        # n2[0] =  n3[1]*n1[2]-n3[2]*n1[1]
        # n2[1] = -n1[2]*n3[0]+n1[0]*n3[2]
        # n2[2] =  n3[0]*n1[1]-n1[0]*n3[1]

        #An = np.append([n1,n2],[n3],0)

        # G = np.mat([
        #     [ n1[0], n1[1], n1[2], 0,     0,     0,     0,     0,     0,     0,     0,     0    ],
        #     [ n2[0], n2[1], n2[2], 0,     0,     0,     0,     0,     0,     0,     0,     0    ],
        #     [ n3[0], n3[1], n3[2], 0,     0,     0,     0,     0,     0,     0,     0,     0    ],
        #     [ 0,     0,     0,     n1[0], n1[1], n1[2], 0,     0,     0,     0,     0,     0    ],
        #     [ 0,     0,     0,     n2[0], n2[1], n2[2], 0,     0,     0,     0,     0,     0    ],
        #     [ 0,     0,     0,     n3[0], n3[1], n3[2], 0,     0,     0,     0,     0,     0    ],
        #     [ 0,     0,     0,     0,     0,     0,     n1[0], n1[1], n1[2], 0,     0,     0    ],
        #     [ 0,     0,     0,     0,     0,     0,     n2[0], n2[1], n2[2], 0,     0,     0    ],
        #     [ 0,     0,     0,     0,     0,     0,     n3[0], n3[1], n3[2], 0,     0,     0    ],
        #     [ 0,     0,     0,     0,     0,     0,     0,     0,     0,     n1[0], n1[1], n1[2]],
        #     [ 0,     0,     0,     0,     0,     0,     0,     0,     0,     n2[0], n2[1], n2[2]],
        #     [ 0,     0,     0,     0,     0,     0,     0,     0,     0,     n3[0], n3[1], n3[2]]
        # ])
        
        # Ke = G.T*Kle*G
        # fe = G.T*fle
        return Kle  
        # if eq == None:
        #     return Kle
        # else:
        #     return Kle,fe


    def ke_matrix(self):
        Kle = self.k_matrix()
        ex = [self.nodes[0].x, self.nodes[1].x]
        ey = [self.nodes[0].y, self.nodes[1].y]
        ez = [self.nodes[0].z, self.nodes[1].z]
        b = np.mat([
                    [ex[1]-ex[0]],
                    [ey[1]-ey[0]],
                    [ez[1]-ez[0]]])

        L = self.L
        n1 = np.asarray(b.T/L).reshape(3,)
        
        eo = np.asmatrix(eo)
        lc = np.asscalar(np.sqrt(eo*eo.T))
        n3 = np.asarray(eo/lc).reshape(3,)
        
        
        n2 = np.array([0.,0.,0.])
        n2[0] =  n3[1]*n1[2]-n3[2]*n1[1]
        n2[1] = -n1[2]*n3[0]+n1[0]*n3[2]
        n2[2] =  n3[0]*n1[1]-n1[0]*n3[1]

        T = np.mat([
            [ n1[0], n1[1], n1[2], 0,     0,     0,     0,     0,     0,     0,     0,     0    ],
            [ n2[0], n2[1], n2[2], 0,     0,     0,     0,     0,     0,     0,     0,     0    ],
            [ n3[0], n3[1], n3[2], 0,     0,     0,     0,     0,     0,     0,     0,     0    ],
            [ 0,     0,     0,     n1[0], n1[1], n1[2], 0,     0,     0,     0,     0,     0    ],
            [ 0,     0,     0,     n2[0], n2[1], n2[2], 0,     0,     0,     0,     0,     0    ],
            [ 0,     0,     0,     n3[0], n3[1], n3[2], 0,     0,     0,     0,     0,     0    ],
            [ 0,     0,     0,     0,     0,     0,     n1[0], n1[1], n1[2], 0,     0,     0    ],
            [ 0,     0,     0,     0,     0,     0,     n2[0], n2[1], n2[2], 0,     0,     0    ],
            [ 0,     0,     0,     0,     0,     0,     n3[0], n3[1], n3[2], 0,     0,     0    ],
            [ 0,     0,     0,     0,     0,     0,     0,     0,     0,     n1[0], n1[1], n1[2]],
            [ 0,     0,     0,     0,     0,     0,     0,     0,     0,     n2[0], n2[1], n2[2]],
            [ 0,     0,     0,     0,     0,     0,     0,     0,     0,     n3[0], n3[1], n3[2]]
        ])
        Ke = T.T*Kle*T
        return Ke

class TensorIsoQuad(Element):    
    def __init__(self, tag, iNode, jNode, E=None, A=None, properties=None,geom='lin'):
        super().__init__(self.ndf, self.ndm, self.force_dict, [iNode, jNode])
        
        if isinstance(properties,dict):
            E, A = properties['E'], properties['A']

class PlaneQuad:
    def __init__(self,ndf):
        self.nn = 4
        self.ndf = ndf 
        self.E = 29.0e3
        self.nu = 0.25

        self.N = np.empty((2,ndf*4),dtype=object)
        self.B = np.empty((3,ndf*4),dtype=object)
        self.problem_type = 'plane-stress'
    
    
    def D(self,problem_type=None):
        if problem_type is None: problem_type = self.problem_type

        nu = self.nu 
        E = self.E
        D = E/(1-nu**2)*np.array([
            [1.0,  nu, 0.0],
            [ nu, 1.0, 0.0],
            [0.0, 0.0, 0.5*(1-nu)]
        ])

    def k_matrix(self,u=None):
        D = self.D()
        B = self.B
        k = np.zeros(())

        for i,Bi in enumerate(B.T):
            for j,Bj in enumerate(B.T):
                f = lambda x,y: Bj(x,y).T@D@Bi(x,y)
                k[j,j] = scipy.integrate.nquad(f,[[-1,1],[-1,1]])[0]

        return k




"""Atomic operations implemented through the JAX Numpy API.

[`jax.numpy`](https://jax.readthedocs.io/en/latest/jax.numpy.html)

"""
from functools import wraps
import numpy as np
# from anon.conf import _BACKEND 

# backend = _BACKEND['ops']
backend = ["jax"]
try:
    from jax.config import config
    config.update("jax_enable_x64", True)
    from jax.numpy import *
except:
    from numpy import *

@wraps(np.genfromtxt)
def genfromtxt(*args, **kwds):
    return asarray(np.genfromtxt(*args,**kwds))

@wraps(np.empty)
def empty(*args,**kwds):
    return np.empty(*args,**kwds)



import functools
import numbers
import numpy as np

#def argstodual(e):
#    def decorator(func):
#        @functools.wraps(func)
#        def wrapper(*args, **kwargs):
#            args = [Dual(arg,e) if isinstance(arg, numbers.Number) else arg for arg in args]
#            for key in kwargs:
#                if isinstance(kwargs[key], numbers.Number):
#                    kwargs[key] = Dual(kwargs[key], e)
#            return func(*args, **kwargs)
#        return wrapper
#    return decorator

def othertodual(e):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, other):
            if isinstance(other, numbers.Number):
                other = Dual(real=other, nvars=self.nvars)
            return func(self, other)
        return wrapper
    return decorator

#def grad(func):
#    @functools.wraps(func)
#    @argstodual(1)
#    def wrapper(*args, **kwargs):
#        return func(*args, **kwargs)
#    return wrapper

def grad(*args):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if varpos is None:
                # Full gradient case
                # Count numerical arguments
                nvars = len([a for a in args if isinstance(a, numbers.Number)])
                nvars += len([key for key in kwargs if isinstance(kwargs[key], numbers.Number)])
                # Count numpy args
                nvars += sum([a.size for a in args if isinstance(a, np.ndarray)])
                nvars += sum([kwargs[key].size for key in kwargs if isinstance(kwargs[key], np.ndarray)])
                newargs = [arg for arg in args]
                i=0
                for k,arg in enumerate(args):
                    if isinstance(arg, numbers.Number):
                        newargs[k] = Dual(arg, nvars=nvars, seedvar=i)
                        i += 1
                    elif isinstance(arg, np.ndarray):
                        # Each element is its own variable
                        numpytodual = np.vectorize(lambda x : Dual(x, nvars=nvars))
                        newargs[k] = numpytodual(arg)
                        flat_iter = newargs[k].flat
                        for elt in range(newargs[k].size):
                            flat_iter[elt].dual[i] = 1
                            i += 1
                for key in kwargs:
                    if isinstance(kwargs[key], numbers.Number):
                        kwargs[key] = Dual(kwargs[key], nvars=nvars, seedvar=i)
                        i += 1
                    elif isinstance(kwargs[key], np.ndarray):
                        numpytodual = np.vectorize(lambda x : Dual(x, nvars=nvars))
                        kwargs[key] = numpytodual(kwargs[key])
                        flat_iter = kwargs[key].flat
                        for elt in range(kwargs[key].size):
                            flat_iter[elt].dual[i] = 1
                            i += 1
                args = newargs
            else:
                #nvars = len(varpos)
                nvars = 0
                for i in varpos:
                    if isinstance(args[i], np.ndarray):
                        nvars += args[i].size
                    else:
                        nvars += 1
                if kwargs:
                    raise Exception("Keyword arguments only supported for full gradient.")
                newargs = [arg for arg in args]
                # Replace the chosen vars in varpos with dual[i]=1
                i = 0
                for k in varpos:
                    if isinstance(args[k], np.ndarray):
                        #numpytodual = np.vectorize(lambda x : Dual(x, nvars=nvars, seedvar=i))
                        #newargs[k] = numpytodual(args[k])
                        # Each element is its own variable
                        numpytodual = np.vectorize(lambda x : Dual(x, nvars=nvars))
                        newargs[k] = numpytodual(args[k])
                        flat_iter = newargs[k].flat
                        for elt in range(newargs[k].size):
                            flat_iter[elt].dual[i] = 1
                            i += 1
                    else:
                        newargs[k] = Dual(args[k], nvars=nvars, seedvar=i)
                        i += 1
                args = newargs
            return func(*args, **kwargs)
        return wrapper
    if len(args)==1 and callable(args[0]):
        # @grad with no arguments - take full gradient
        varpos = None
        return decorator(args[0])
    elif len(args)==0:
        # grad() with no arguments
        varpos = None
    else:
        varpos = args[0]
        if isinstance(varpos, numbers.Number):
            varpos = [varpos]
    return decorator


def dot(x, y):
    ret = Dual(0.0)
    for i in range(len(x)):
        ret += x[i]*y[i]
    return ret

class Dual:
    def __init__(self, real=0, dual=None, nvars=None, seedvar=None):
        self.real = real
        if dual is None and nvars is None:
            self.dual = np.zeros(1)
            self.nvars = 1
        elif nvars is None:
            self.dual = np.array(dual, dtype=np.float64, ndmin=1)
            self.nvars = len(self.dual)
        elif dual is None:
            self.dual = np.zeros(nvars)
            self.nvars = nvars
        else:
            self.dual = np.array(dual, dtype=np.float64, ndmin=1)
            self.nvars = nvars
        if seedvar is not None:
            self.dual[seedvar] = 1
    def __add__(self, other):
        otherreal = getattr(other, 'real', other)
        otherdual = getattr(other, 'dual', None)
        ret = Dual(self.real + otherreal, self.dual)
        if otherdual is not None:
            ret.dual += otherdual
        return ret
    def __sub__(self, other):
        otherreal = getattr(other, 'real', other)
        otherdual = getattr(other, 'dual', None)
        ret = Dual(self. real - otherreal, self.dual)
        if otherdual is not None:
            ret.dual -= otherdual
        return ret
    def __mul__(self, other):
        otherreal = getattr(other, 'real', other)
        otherdual = getattr(other, 'dual', None)
        ret = Dual(self.real * otherreal, self.dual * otherreal)
        if otherdual is not None:
            ret.dual += self.real*otherdual
        return ret
    def __truediv__(self, other):
        otherreal = getattr(other, 'real', other)
        otherdual = getattr(other, 'dual', None)
        ret = Dual(self. real / otherreal, self.dual / otherreal)
        if otherdual is not None:
            ret.dual -= self.real*otherdual/(otherreal*otherreal)
        return ret
    # object.__floordiv__(self, other)
    # object.__mod__(self, other)
    # object.__divmod__(self, other)
    def __pow__(self, other, *modulo):
        if modulo:
            return NotImplemented
        if isinstance(other, int):
            m = other
            negative = (m<0)
            m = abs(m)
            ret = Dual(1, nvars=self.nvars)
            for _ in range(m):
                ret *= self
            if negative:
                ret = 1/ret
            return ret
        else:
            return Dual.exp(other*Dual.log(self))
    # object.__lshift__(self, other)
    # object.__rshift__(self, other)
    # object.__and__(self, other)
    # object.__xor__(self, other)
    # object.__or__(self, other)

    def __radd__(self, other):
        return self.__add__(other)
    def __rsub__(self, other):
        return -self.__sub__(other)
    def __rmul__(self, other):
        return self.__mul__(other)
    def __rtruediv__(self, other):
        return Dual(1,nvars=self.nvars)/self.__truediv__(other)
    # object.__rfloordiv__(self, other)
    # object.__rmod__(self, other)
    # object.__rdivmod__(self, other)
    def __rpow__(self, other, *modulo):
        if modulo:
            return NotImplemented
        x = Dual(other, nvars=self.nvars)
        return x.__pow__(self)
    # object.__rlshift__(self, other)
    # object.__rrshift__(self, other)
    # object.__rand__(self, other)
    # object.__rxor__(self, other)
    # object.__ror__(self, other)

    def __iadd__(self, other):
        otherreal = getattr(other, 'real', other)
        otherdual = getattr(other, 'dual', None)
        self.real += otherreal
        if otherdual is not None:
            self.dual += other.dual
        return self
    def __isub__(self, other):
        otherreal = getattr(other, 'real', other)
        otherdual = getattr(other, 'dual', None)
        self.real -= otherreal
        if otherdual is not None:
            self.dual -= other.dual
        return self
    # object.__imul__(self, other)
    # object.__itruediv__(self, other)
    # object.__ifloordiv__(self, other)
    # object.__imod__(self, other)
    # object.__ipow__(self, other[, modulo])
    # object.__ilshift__(self, other)
    # object.__irshift__(self, other)
    # object.__iand__(self, other)
    # object.__ixor__(self, other)
    # object.__ior__(self, other)

    def __neg__(self):
        return Dual(-self.real, -self.dual)
    # object.__pos__(self)
    # object.__abs__(self)
    # object.__invert__(self)

    @othertodual(0)
    def __lt__(self, other):
        return self.real < other.real
    @othertodual(0)
    def __le__(self, other):
        return self.real <= other.real
    @othertodual(0)
    def __gt__(self, other):
        return self.real > other.real
    @othertodual(0)
    def __ge__(self, other):
        return self.real >= other.real

    def __str__(self):
        return "Dual({},  {})".format(self.real, self.dual)
    def __repr__(self):
        return self.__str__()

    # f(a + be) = f(a) + b fprime(a) e

    def exp(x):
        expa = np.exp(x.real)
        ret = Dual(expa, x.dual*expa)
        return ret
    
    def log(x):
        ret = Dual(np.log(x.real), x.dual / x.real)
        return ret
    
    def sin(x):
        ret = Dual(np.sin(x.real), x.dual * np.cos(x.real))
        return ret
    
    def cos(x):
        ret = Dual(np.cos(x.real), -x.dual * np.sin(x.real))
        return ret
    
    def sqrt(x):
        sqrta = np.sqrt(x.real)
        ret = Dual(sqrta, x.dual * 0.5/sqrta)
        return ret

#def minimise(f, x0, alpha = 1e-1, maxits = 1000, tolerance=1e-6, variables=None, verbose=False):
#    fgrad = grad(variables)(f)
#    x0 = np.array(x0)
#    zerostep = 0.0*x0 # Maintain shape of each element of x0
#    tolerance_sq = tolerance**2
#    converged = False
#    for it in range(maxits):
#        func = fgrad(*x0)
#        gradient = func.dual
#        if variables is None:
#            step = alpha*gradient
#        else:
#            # Only step in specified variables
#            step = zerostep
#            step[variables] = alpha*gradient
#        x0 = x0 - step
#        stepsizesq = np.dot(gradient, gradient)
#        #print(gradient)
#        if stepsizesq<tolerance_sq:
#            converged = True
#            break
#        if verbose and it%5==0:
#            print("Iter {}: fval = {}".format(it, func.real))
#    if not converged:
#        print("Warning: didn't converge in {} iterations".format(maxits))
#    result = fgrad(*x0)
#    return x0, result.real, result.dual




"""
# Matrix API

This module provides functions and classes for constructing various structural analysis matrices. 

"""

import numpy as np
import pandas as pd
import scipy.linalg
import matplotlib.pyplot as plt

try:
    import emme
except:
    pass
# from emme.utilities import Structural_Matrix, Structural_Vector, transfer_vars


settings = {
    "DATAFRAME_LATEX": True,
}

subscripts = {
    'part': 'subscript',
     None: '',
    'initial': '0',
    'continuous':'c',
    'primary':'i',
    'reactions':'d',
    'elem-load':'w',
    'free': 'f',
}

def del_zeros(mat):
    delrows = np.where(~mat.any(axis=1))[0]
    delcols = np.where(~mat.any(axis=0))[0]

    newM = np.delete(mat, delcols, axis=1)
    newM = np.delete(newM, delrows, axis=0)
    return newM

def elem_dofs(Elem):
    dofs = []
    for node in Elem.nodes:
        dofs.extend(node.dofs)
    return dofs

def transfer_vars(item1, item2):
    for key, value in item1.__dict__.items():
        item2.__dict__[key] = value

def Localize(U_vector, P_vector, model=None):
    if model is None: model = U_vector.model
    A =  A_matrix(model)
    Q0 = Q0_vector(model)
    Ks = Ks_matrix(model)

    V = A.f @ U_vector.f
    Q  = Ks@V + Q0
    return V, Q



class Structural_Vector(np.ndarray):
    column_data = ["vector"]
    row_data = None
    subs = [None]
    tag = 'Vector'
    def __new__(cls, mat):
        mat = mat
        return np.asarray(mat).view(cls)

    def __init__(self, mat):
        self.tag = None

    def _repr_html_(self):
        try:
            out = self.df.to_html()
            return out
        except: 
            return self

    def __add__(self, other):
        if isinstance(other, type(self)):
            out = np.add(self, other).view(type(self))
            transfer_vars(self, out)
        else:
            out = super().__add__(other)
        return out
    
    def get(self, key):
        idx = np.array([i for i,j in enumerate(self.row_data) if str(j) == key], dtype=int)
        # row = self.row_data.index(component)
        return self[idx]

    def set_item(self, key, value):
        idx = np.array([i for i,j in enumerate(self.row_data) if str(j) == key], dtype=int)
        self[idx] = value

    def rows(self, component):
        idx = np.where(np.isin(self.row_data,component))[0]
        newV = self[idx]
        newV.row_data = np.array(self.row_data)[idx]
        newV.model = self.model
        return newV

    @property
    def df(self):
        row_data = ['$'+str(tag)+'$' for tag in self.row_data]
        header = '$'+self.tag+'_{{'
        try:
            for sub in self.subs:
                header += subscripts[sub]
        except: pass
        print
        header += '}}$'

        return pd.DataFrame(np.around(self,14), index=row_data, columns=[header])

    @property
    def symb(self):
        var = []
        for eid in self.row_data:
            var.append(sp.symbols(self.tag+eid))
        return sp.Matrix(var)
    
    @property
    def disp(self):
        return sp.Matrix(self)
     
class Structural_Matrix(np.ndarray):
    column_data = None
    row_data = None
    c_ridx = None # row indexes for use in .c method
    c_cidx = None # column indexes for use in .c method
    tag = None

    def __new__(cls, mat):
        mat = mat
        return np.asarray(mat).view(cls)

    def __init__(self, mat):
        self.tag = None

    def __matmul__(self, other):
        if isinstance(other, Structural_Matrix):
            out = np.matmul(self,other).view(Structural_Matrix)
            transfer_vars(self,out)
            out.column_data = other.column_data
        
        elif isinstance(other, Structural_Vector):
            out = np.matmul(self,other).view(Structural_Vector)
            out.row_data = self.row_data
            # out.column_data = ['.']
        else:
            out = np.matmul(self,other).view(Structural_Vector)
            out.row_data = self.row_data
            # out.column_data = ['.']
        return out

    def __add__(self, other):
        out = np.add(self,other)
        if (isinstance(other, Structural_Matrix)
            or isinstance(other, Structural_Vector)):
            out.row_data = self.row_data
        return out

    
    def __truediv__(self, other):
        out = np.ndarray.__truediv__(self, other) 
        if (isinstance(other, float) 
            or isinstance(other, Structural_Matrix)
            or isinstance(other, Structural_Vector)):
            out = np.ndarray.__truediv__(self, other).view(Structural_Matrix)
            transfer_vars(self, out)
        else:
            out = np.ndarray.__truediv__(self, other)
        return out
        
    def _repr_html_(self):
        try: 
            df = self.df
            return df.to_html()
        except:
            try:
                return pd.DataFrame(self).to_html()
            except:
                pass

    @property
    def disp(self):
        return sp.Matrix(self)
    
    @property
    def df(self):
        if settings['DATAFRAME_LATEX']:
            row_data = ['$'+str(tag)+'$' for tag in self.row_data]
            column_data = ['$'+str(tag)+'$' for tag in self.column_data]
        else:
            row_data = [str(i) for i in self.row_data]
            column_data = [str(i) for i in self.column_data]
        return pd.DataFrame(np.around(self,5), index=row_data, columns=column_data)

    @property
    def inv(self):
        mat = np.linalg.inv(self)
        transfer_vars(self, mat)
        mat.row_data = self.column_data
        mat.column_data = self.row_data
        return mat
    
    @property
    def rank(self):
        """Return the rank of a matrix"""
        return np.linalg.matrix_rank(self)
    
    @property
    def lns(self):
        """Return a basis for the left nullspace of a matrix."""
        return scipy.linalg.null_space(self.T)
    
    @property
    def nls(self):
        """return a basis for the nullspace of matrix."""
        return scipy.linalg.null_space(self)
    
    @property
    def ker(self):
        "Return a basis for the kernel (nullspace) of a matrix."
        kernel = scipy.linalg.null_space(self) 
        ker = Structural_Matrix(kernel)
        transfer_vars(self,ker)
        ker.row_data = self.column_data 
        ker.column_data = [str(i+1) for i in range(len(ker[0]))]
        return ker

    def lu(self):
        
        return scipy.linalg.lu(self)

    @property
    def c(self):
        delcols = self.c_cidx
        delrows = self.c_ridx

        newM = np.delete(self, delrows, axis=0).view(type(self))
        if delcols: newM = np.delete(newM, delcols, axis=1).view(type(self))
        
        transfer_vars(self, newM)

        if delcols: newM.column_data = list(np.delete(self.column_data, delcols))

        newM.row_data = list(np.delete(self.row_data, delrows))
        return newM

    def round(self, num):
        newM = np.around(self, num).view(Structural_Matrix)
        transfer_vars(self,newM)
        return newM

    def remove(self, component):
        """Remove items by looking up column_data/row_data"""
        if type(component) is list:
            for item in component:
                if item in self.column_data: 
                    delcol = self.column_data.index(item)
                    newM = np.delete(self, delcol, axis=1).view(type(self))
                    transfer_vars(self,newM)
                    newM.column_data = list(np.delete(self.column_data, delcol))
                else:
                    delrow = self.row_data.index(item)
                    newM = np.delete(self, delrow, axis=0).view(type(self))
                    transfer_vars(self, newM)
                    # newM.column_data = self.column_data
                    # newM.model = self.model
                    newM.row_data = list(np.delete(self.row_data, delrow))

        else:
            item = component
            if item in self.column_data: 
                delcol = self.column_data.index(item)
                newM = np.delete(self, delcol, axis=1).view(type(self))
                newM.row_data = self.row_data
                newM.model = self.model
                newM.column_data = list(np.delete(self.column_data, delcol))
                # try: newM.rel = self.rel
                # except: pass
            else:
                delrow = self.row_data.index(item)
                newM = np.delete(self, delrow, axis=0).view(type(self))
                newM.column_data = self.column_data
                newM.model = self.model
                newM.row_data = list(np.delete(self.row_data, delrow))
                # try: newM.rel = self.rel
                # except: pass
        return newM

    def get(self, row_name, col_name):
        idxr = np.where(self.row_data == row_name)
        idxc = np.where(self.column_data == col_name)
        return self[idxr, idxc]


    def rows(self, component):
        rows = [self.row_data.index(item) for item in component]
        newM = self[rows,:]
        newM.model = self.model
        newM.column_data = self.column_data
        newM.row_data = list(np.array(self.row_data)[rows])
        return newM

    def del_zeros(self):
        """Delete rows and columns of a matrix with all zeros"""
        delrows = np.where(~self.any(axis=1))[0]
        delcols = np.where(~self.any(axis=0))[0]

        newM = np.delete(self, delcols, axis=1).view(type(self))
        newM = np.delete(newM, delrows, axis=0).view(type(self))
        transfer_vars(self, newM)

        newM.column_data = list(np.delete(self.column_data, delcols))
        newM.row_data = list(np.delete(self.row_data, delrows))
        return newM
        
    def add_cols(self, component):
        if "colinear" in component:
            vertical = [elem for elem in self.model.elems if elem.Dx==0.0]
            other = set({})
            for elem in self.model.elems:
                try: other.add(elem.Dy/elem.Dx)
                except ZeroDivisionError: pass
            return vertical, other
        if type(component) is list:
            #ASSUMES component IS A LIST OF COLUMN INDICES
            delcols = [self.column_data.index(item) for item in component[1:len(component)]]
            i0 = self.column_data.index(component[0])
            newM = np.delete(self, delcols, axis=1).view(type(self))
            for col in delcols:
                newM[:,i0] += self[:,col]
            newM.row_data = self.row_data
            newM.model = self.model
            newM.column_data = list(np.delete(self.column_data, delcols))
            # try: newM.rel = self.rel
            # except: pass
            return newM

    def add_rows(self, component):
        if type(component) is list:
            delrows = [self.row_data.index(item) for item in component[1:len(component)]]
            i0 = self.row_data.index(component[0])
            newA = np.delete(self, delrows, axis=0).view(type(self))
            for row in delrows:
                newA[i0,:] += self[row,:]
            newA.column_data = self.column_data
            newA.model = self.model
            newA.row_data = list(np.delete(self.row_data, delrows))
            return newA

class row_vector (Structural_Vector):
    def __new__(cls, Matrix):
        V = np.zeros((len(Matrix.row_data)))
        return np.asarray(V).view(cls)

    def __init__(self, Matrix):
        self.tag = "Y"
        self.matrix = Matrix 
        self.row_data = Matrix.row_data
    
class column_vector (Structural_Vector):

    def __new__(cls, Matrix, Vector=None):
        V = np.zeros((len(Matrix.column_data)))
        return np.asarray(V).view(cls)

    def __init__(self, Matrix, Vector=None):
        self.tag = "X"
        self.matrix = Matrix 
        self.row_data = Matrix.column_data
        if Vector is not None:
            for key in Vector.row_data:
                self.set_item(key, Vector.rows([key]))


class Static_matrix (Structural_Matrix):
    """B_MATRIX static matrix of structural model with 2d/3d truss and 2d frame elements
    the function forms the static matrix B for all degrees of freedom and
    all basic forces of the structural model specified in data structure MODEL;
    the function is currently limited to 2d/3d truss and 2d frame elements
    
    Parameters
    ---------------

    model: emme.Model object
    
    Partitions
    =========================================================================================
 
    - B.f  : nf x ntq

    - B.c  : nf x nq

    - B.fc : nf x nq

    - B.i  : ni x nq

    - B.x  : nx x nq

    where:

    - ni: number of primary (non-redundant) forces.
    - nq: number of total, continuous forces.
    - nx: number of redundant forces.

    """

    ranges = {
        'f': 'free dof rows',
        'i': 'primary force columns',
        'x': 'redundant force columns',
        'c': 'continuous force columns (removes releases)',
        'd': 'reaction force columns',
    }

    def __new__(cls, model, matrix=None, rng=None):
        fullnq = sum([len(elem.rel) for elem in model.elems])
        B = np.zeros((model.nt,fullnq))
        ci = 0
        for elem in model.elems:
            dofs = elem.dofs
            bg = elem.bg_matrix()
            nq = np.size(bg,1)
            for j,dof in enumerate(dofs):
                B[int(dof)-1,ci:ci+nq] = bg[j,:]
            ci = ci+nq
        input_array = B
        return np.asarray(input_array).view(cls)

    def __init__(self, model, matrix=None, rng=None):
        if rng is None:
            self.rng = None
        self.model = model
        self.row_data = np.array([str(dof) for dof in range(1, model.nt+1)])
        self.column_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
        if matrix is not None:
            fullnq = sum([len(elem.rel) for elem in model.elems])
            self[:,:] = np.zeros((model.nt,fullnq))
            for idxr, rw in enumerate(self.row_data):
                if rw in matrix.row_data:
                    for idxc, cl in enumerate(self.column_data):
                        if cl in matrix.column_data:
                            self[idxr,idxc] = matrix.get(rw,cl)

    def __matmul__(self, Vector):
        if type(Vector) is nForce_vector:
            vect = np.matmul(self, Vector).view(iForce_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif type(Vector) is iForce_vector:
            vect = np.matmul(self, Vector).view(nForce_vector) 
            vect.row_data = self.row_data
            vect.matrix = self
        else:
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect
  #<# 
    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.model.nf]
        
        newB = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self, newB)
        # newB.model = self.model
        newB.row_data = list(np.delete(self.row_data, delrows))
        # newB.column_data = self.column_data
        return newB

    @property
    def i(self):
        """Removes rows of B_matrix corresponding to primary (non-redundant) forces"""
        Bf = self.f
        idx_i = self.model.idx_i
        newB = Bf[:,idx_i]
        # newB = Bf[idx_i,:]
        transfer_vars(Bf, newB)
        newB.column_data = Bf.column_data[idx_i]
        return newB
    
    @property
    def d(self):
        """Removes rows corresponding to free dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) <= self.model.nf]
        newB = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self,newB)
        newB.row_data = list(np.delete(self.row_data, delrows))
        return newB

    @property
    def c(self):

        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = np.delete(Bf, delcols, axis=1).view(type(self))
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        return newB

    @property
    def o(self):
        """Remove columns corresponding to element force releases, then delete zeros"""
        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = np.delete(Bf, delcols, axis=1).view(type(self))
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        newB = newB.del_zeros()
        return newB

    @property
    def fc(self):
        return self.f.c

    @property
    def x(self):
        """Removes rows of B_matrix corresponding to primary (non-redundant) forces"""
        idx_x = self.model.idx_x
        newB = self[:,idx_x]
        transfer_vars(self, newB)
        newB.column_data = self.column_data[idx_x]
        return newB

    @property
    def barxi(self):
        Bx = self.f.x
        Bbarxi = self.bari @ -Bx
        Bbarxi.column_data = Bx.column_data
        return Bbarxi

    @property
    def barx(self):
        nQ = len(self.column_data)
        nx = len(self.model.redundants)

        Bbarxi = self.barxi

        Bbarx = Structural_Matrix(np.zeros((nQ,nx)))
        transfer_vars(self, Bbarx)
        Bbarx.column_data = Bbarxi.column_data
        Bbarx.row_data = self.column_data

        for idxc, cl in enumerate(Bbarx.column_data):
            for idxr, rw in enumerate(Bbarx.row_data):
                if rw in Bbarxi.row_data:
                    Bbarx[idxr,idxc] = Bbarxi.get(rw,cl)
                elif cl==rw:
                        Bbarx[idxr, idxc] = 1.

        return Bbarx

    @property
    def bari(self):
        return self.i.del_zeros().inv

    @property
    def c0(self):
        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = Bf 
        for col in delcols:
            newB[:,col] = [0.]*len(Bf[:,0])
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        return newB

class nStatic_matrix(Structural_Matrix):
    ranges = {
        'f': 'free dof rows',
        'i': 'primary force columns',
        'x': 'redundant force columns',
        'c': 'continuous force columns (removes releases)',
        'd': 'reaction force columns',}
    def __new__(cls, arry, model, rcdata):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, rcdata):
        self.model = model
        self.row_data = rcdata[0]
        self.column_data = rcdata[1]
        
    def __matmul__(self, Vector):
        if type(Vector) is nForce_vector:
            vect = np.matmul(self, Vector).view(iForce_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif isinstance(Vector, iForce_vector):
            vect = np.matmul(self, Vector).view(nForce_vector) 
            vect.row_data = self.row_data
            vect.matrix = self
        else:
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect

    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.model.nf]
        
        newB = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self, newB)
        newB.row_data = np.delete(self.row_data, delrows)
        return newB
        
    @property
    def i(self):
        """Removes rows of B_matrix corresponding to redundant forces"""
        # reducedB = self.f.c.del_zeros()
        reducedB = self.f.c
        rdts = reducedB.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        delcols = [reducedB.column_data.index(tag) for tag in tags]
        newB = np.delete(reducedB, delcols, axis=1).view(Static_matrix)
        transfer_vars(reducedB, newB)
        newB.column_data = list(np.delete(reducedB.column_data, delcols))
        return newB
    
    @property
    def d(self):
        """Removes rows corresponding to free dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) <= self.model.nf]
        newB = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self,newB)
        newB.row_data = list(np.delete(self.row_data, delrows))
        return newB

    @property
    def c(self):
        """Removes columns corresponding to element hinges/releases"""
        Af = self.f
        idx_c = self.model.idx_c
        newA = Af[:,idx_c]
        transfer_vars(Af, newA)
        newA.column_data = Af.column_data[idx_c]
        return newA

    @property
    def o(self):
        """Remove columns corresponding to element force releases, then delete zeros"""
        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = np.delete(Bf, delcols, axis=1).view(type(self))
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        newB = newB.del_zeros()
        return newB

    @property
    def fc(self):
        return self.f.c
    
    @property
    def x(self):
        """Removes rows of B_matrix corresponding to primary (non-redundant) forces
        
        """
        rdts = self.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        cols = [self.column_data.index(tag) for tag in tags]
        newB = self[:,cols]
        transfer_vars(self, newB)
        newB.column_data = np.array([self.column_data[col] for col in cols])
        # newB.row_data = self.row_data
        # newB.model = self.model
        return newB

    @property
    def barxi(self):
        Bx = self.f.x
        Bbarxi = self.bari @ -Bx
        Bbarxi.column_data = Bx.column_data
        return Bbarxi

    @property
    def barx(self):
        nQ = len(self.column_data)
        nx = len(self.model.redundants)

        Bbarxi = self.barxi

        Bbarx = Structural_Matrix(np.zeros((nQ,nx)))
        transfer_vars(self, Bbarx)
        Bbarx.column_data = Bbarxi.column_data
        Bbarx.row_data = self.column_data
        for idxc, cl in enumerate(Bbarx.column_data):
            for idxr, rw in enumerate(Bbarx.row_data):
                if rw in Bbarxi.row_data:
                    Bbarx[idxr,idxc] = Bbarxi.get(rw,cl)
                elif cl==rw:
                        Bbarx[idxr, idxc] = 1.

        return Bbarx

    @property
    def bari(self):
        return self.i.del_zeros().inv

    @property
    def c0(self):
        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = Bf
        for col in delcols:
            newB[:,col] = [0.]*len(Bf[:,0])
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        # newB.row_data = self.row_data
        # newB.rel = self.rel
        return newB

def B_matrix(model, matrix=None, rng=None):
    """Returns a Static_matrix object"""
    return Static_matrix(model, matrix, rng)

def nB_matrix(model, rng=None):
    """Returns a Static_matrix object"""
    fullnq = sum([len(elem.rel) for elem in model.elems])
    B = np.zeros((model.nt, fullnq))
    ci = 0
    for elem in model.elems:
        eid = elem.dofs
        bg = elem.bg_matrix()
        nq = np.size(bg,1)
        for j,eidi in enumerate(eid):
            B[int(eidi)-1,ci:ci+nq] = bg[j,:]
        ci = ci+nq
    input_array = B
    return np.asarray(input_array).view(nStatic_matrix)
    # matrix =  np.asarray(input_array)
    # return nStatic_matrix(model, matrix, rng)

def Bh_matrix(model):
    """Returns a Static_matrix object"""
    fullnq = sum([len(elem.rel) for elem in model.elems])
    B = np.zeros((model.nt, fullnq))
    ci = 0
    for elem in model.elems:
        eid = elem.dofs
        bg = elem.bg_matrix(Roption=True)
        nq = np.size(bg,1)
        for j,eidi in enumerate(eid):
            B[int(eidi)-1,ci:ci+nq] = bg[j,:]
        ci = ci+nq
    row_data = np.array([str(dof) for dof in range(1, model.nt+1)])
    column_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
    rcdata = (row_data, column_data)
    return nStatic_matrix(B, model, rcdata)

class Kinematic_matrix (Structural_Matrix):
    """Class for the kinematic matrix of a structural model with 2d/3d truss and 2d frame elements
    the function forms the kinematic matrix A for all degrees of freedom and
    all element deformations of the structural model specified in data structure MODEL
    the function is currently limited to 2d/3d truss and 2d frame elements
    
    Returns
    ---------

    Kinematic matrix

    """

    ranges = {
        'f': 'free dof columns',
        'i': 'primary force/deformation rows',
        'x': 'redundant force/deformation rows',
        'd': 'reaction force/deformation rows',
        'c': 'continuous (hingeless) force/deformation rows'
    }

    def __new__(cls, model, matrix=None,rng=None):
        A  = np.zeros((sum(model.nv),model.nt))
        ri = 0
        for elem in model.elems:
            eid = elem.dofs
            ag = elem.ag()
            nv = len(elem.v)
            for j, eidi in enumerate(eid):
                A[ri:ri+nv, int(eidi)-1] = ag[:,j]
            ri = ri+nv
        input_array = A
        return np.asarray(input_array).view(cls)

    def __init__(self, model, matrix=None,rng=None):
        if rng is None:
            self.rng = None
        self.model = model
        self.column_data = np.array([str(dof) for dof in range(1,model.nt+1)])
        self.row_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.v.keys()])
        self.basic_deformations = np.array([v for elem in model.elems for v in elem.basic_deformations])
        
        self.idx_h = []
        if matrix is not None:
            fullnq = sum([len(elem.rel) for elem in model.elems])
            self[:,:] = np.zeros((model.nt,fullnq))
            for idxr, rw in enumerate(self.row_data):
                if rw in matrix.row_data:
                    for idxc, cl in enumerate(self.column_data):
                        if cl in matrix.column_data:
                            self[idxr,idxc] = matrix.get(rw,cl)

    def __matmul__(self, Vector):
        if isinstance(Vector, Deformation_vector):
            # print('a')
            vect = np.matmul(self, Vector).view(Displacement_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif isinstance(Vector, Displacement_vector):
            vect = np.matmul(self, Vector).view(Deformation_vector) 
            vect.row_data = self.row_data
            vect.matrix = self
        else:
            vect = Structural_Matrix.__matmul__(self, Vector)

        return vect


    def combine(self, component):
        if "colinear" in component:
            vertical = [elem for elem in self.model.elems if elem.Dx==0.0]
            other = set({})
            for elem in self.model.elems:
                try: other.add(elem.Dy/elem.Dx)
                except ZeroDivisionError: pass
            return vertical, other

        if type(component) is list:
            ## TO BE DEPRECATED
            #ASSUMES component IS A LIST OF COLUMN INDICES
            delcols = [self.column_data.index(item) for item in component[1:len(component)]]
            i0 = self.column_data.index(component[0])
            newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
            for col in delcols:
                newA[:,i0] += self[:,col]
            newA.row_data = self.row_data
            newA.model = self.model
            newA.column_data = list(np.delete(self.column_data, delcols))
            # newA.rel = self.rel
            return newA
    
    @property
    def f(self):
        """Removes columns corresponding to fixed dofs"""
        delcols = [idx for idx, dof in enumerate(self.column_data) if int(dof) > self.model.nf]
        newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
        transfer_vars(self, newA)
        newA.column_data = list(np.delete(self.column_data, delcols))
        return newA

    @property
    def i(self):
        """Removes rows corresponding to redundant forces"""
        Afc = self.f.c
        rdts = self.model.redundants
        tags = [q.elem.tag +'_'+ str(q.nature) for q in rdts]
        delrows = [Afc.row_data.index(tag) for tag in tags]
        newA = np.delete(Afc, delrows, axis=0).view(Kinematic_matrix)
        transfer_vars(Afc,newA)
        newA.row_data = list(np.delete(Afc.row_data, delrows))
        return newA

    @property
    def d(self):
        """Removes columns corresponding to free dofs"""
        delcols = [idx for idx, dof in enumerate(self.column_data) if int(dof) <= self.model.nf]
        newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
        transfer_vars(self,newA)
        newA.column_data = list(np.delete(self.column_data, delcols))
        return newA


    @property
    def c(self):
        """Removes rows corresponding to element hinges/releases"""
        Af = self.f
        idx_c = self.model.idx_c
        newA = Af[idx_c,:]
        transfer_vars(Af, newA)
        newA.row_data = Af.row_data[idx_c]
        return newA
    
    @property
    def c0(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = np.where(np.isin(Af.row_data,tags))
        # delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA
    
    @property
    def o(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        newA = newA.del_zeros()
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA
    
    @property
    def e(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        newA = newA.del_zeros()
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA


class nKinematic_matrix (Structural_Matrix):
    """Class for the kinematic matrix of a structural model with 2d/3d truss and 2d frame elements
    the function forms the kinematic matrix A for all degrees of freedom and
    all element deformations of the structural model specified in data structure MODEL 
    the function is currently limited to 2d/3d truss and 2d frame elements
    
    Returns
    ---------
    Kinematic matrix

    """

    ranges = {
        'f': 'free dof columns',
        'i': 'primary force/deformation rows',
        'x': 'redundant force/deformation rows',
        'd': 'reaction force/deformation rows',
        'c': 'continuous (hingeless) force/deformation rows'
    }

    def __new__(cls, model, matrix=None,rng=None):
        A  = np.zeros((sum(model.nv),model.nt))
        ri = 0
        for elem in model.elems:
            eid = elem.dofs
            ag = elem.ag()
            nv = len(elem.v)
            for j, eidi in enumerate(eid):
                A[ri:ri+nv, int(eidi)-1] = ag[:,j]
            ri = ri+nv
        input_array = A
        return np.asarray(input_array).view(cls)

    def __new__(cls, arry, model, rcdata):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, rcdata):
        self.model = model
        self.row_data = rcdata[0]
        self.column_data = rcdata[1]
        self.basic_deformations = np.array([v for elem in model.elems for v in elem.basic_deformations])
        
        self.idx_h = []


    def __matmul__(self, Vector):
        if isinstance(Vector, Deformation_vector):
            # print('a')
            vect = np.matmul(self, Vector).view(Displacement_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif isinstance(Vector, Displacement_vector):
            # print('b')
            vect = np.matmul(self, Vector).view(Deformation_vector)
            # vect = np.matmul(self, Vector).view(Structural_Vector)  
            vect.row_data = self.row_data
            vect.matrix = self
        else:
            # print('else')
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect

    def combine(self, component):
        if "colinear" in component:
            vertical = [elem for elem in self.model.elems if elem.Dx==0.0]
            other = set({})
            for elem in self.model.elems:
                try: other.add(elem.Dy/elem.Dx)
                except ZeroDivisionError: pass
            return vertical, other

        if type(component) is list:
            ## TO BE DEPRECATED
            #ASSUMES component IS A LIST OF COLUMN INDICES
            delcols = [self.column_data.index(item) for item in component[1:len(component)]]
            i0 = self.column_data.index(component[0])
            newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
            for col in delcols:
                newA[:,i0] += self[:,col]
            newA.row_data = self.row_data
            newA.model = self.model
            newA.column_data = list(np.delete(self.column_data, delcols))
            # newA.rel = self.rel
            return newA
    
    @property
    def f(self):
        """Removes columns corresponding to fixed dofs"""
        delcols = [idx for idx, dof in enumerate(self.column_data) if int(dof) > self.model.nf]
        newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
        transfer_vars(self, newA)
        newA.column_data = list(np.delete(self.column_data, delcols))
        return newA

    @property
    def i(self):
        """Removes rows corresponding to redundant forces"""
        Afc = self.f.c
        rdts = self.model.redundants
        tags = [q.elem.tag +'_'+ str(q.nature) for q in rdts]
        delrows = [Afc.row_data.index(tag) for tag in tags]
        newA = np.delete(Afc, delrows, axis=0).view(Kinematic_matrix)
        transfer_vars(Afc,newA)
        newA.row_data = list(np.delete(Afc.row_data, delrows))
        return newA

    @property
    def d(self):
        """Removes columns corresponding to free dofs"""
        delcols = [idx for idx, dof in enumerate(self.column_data) if int(dof) <= self.model.nf]
        newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
        transfer_vars(self,newA)
        newA.column_data = list(np.delete(self.column_data, delcols))
        return newA


    @property
    def c(self):
        """Removes rows corresponding to element hinges/releases"""
        Af = self.f
        idx_c = self.model.idx_c
        newA = Af[idx_c,:]
        transfer_vars(Af, newA)
        newA.row_data = Af.row_data[idx_c]
        return newA
    
    @property
    def c0(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA
    
    @property
    def o(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        newA = newA.del_zeros()
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA
    
    @property
    def e(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        newA = newA.del_zeros()
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA

def A_matrix(Domain, matrix=None):
    """Returns a Kinematic_matrix object"""
    return Kinematic_matrix(Domain,matrix)

class Flexibility_matrix (Structural_Matrix):
    """
    Parameters
    =========================================================================================
    model
    
    -----------------------------------------------------------------------------------------
    """
    tag = 'F'
    def __new__(cls, model, Roption=True):
        f  = np.array([elem.f_matrix() for elem in model.elems])
        Fs = scipy.linalg.block_diag(*f)


        B = Static_matrix(model)
        Fsr = emme.utilities.del_zeros(Fs)

        F = B.bari.T@(Fsr@B.bari)
        
        input_array = np.asarray(F).view(cls)
        input_array.model = model
        input_array.column_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
        input_array.row_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
        # input_array.rel = [rel for elem in model.elems for rel in elem.rel.values()]

        # # input_array.xx = np.asarray(Fxx).view(cls)
        # input_array.xx.model = model
        # input_array.xx.row_data = B.x.column_data
        # input_array.xx.column_data = B.x.column_data
        # # input_array.xx.rel = [rel for elem in model.elems for rel in elem.rel.values()]


        input_array.s = np.asarray(Fs).view(cls)
        input_array.s.model = model
        input_array.s.column_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
        input_array.s.row_data =    np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
        # input_array.s.rel = [rel for elem in model.elems for rel in elem.rel.values()]

        return input_array

    def __init__(self, model, Roption=True):
        # self.model = model 
        # self.column_data = [elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()]
        # self.row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()]
        # self.rel = [rel for elem in model.elems for rel in elem.rel.values()]
        pass

    def __matmul__(self, Vector):
        if type(Vector) is Deformation_vector:
            vect = np.matmul(self, Vector).view(iForce_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif type(Vector) is iForce_vector:
            vect = np.matmul(self, Vector).view(Deformation_vector)
            # vect.part = 'continuous'
            vect.row_data = self.row_data
            vect.matrix = self

        else:
            vect = Structural_Matrix.__matmul__(self, Vector)

        return vect

    @property 
    def c(self):
        """Removes rows corresponding to element hinges/releases"""
        rows = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        newF = np.delete(self, rows, axis=0).view(type(self))
        newF = np.delete(newF, rows, axis=1).view(type(self))
        newF.row_data = list(np.delete(self.row_data, rows))
        newF.column_data = list(np.delete(self.column_data, rows))
        newF.model = self.model
        newF.tag = self.tag + "_c"
        return newF

def F_matrix(Domain):
    """Returns a Flexibility_matrix object"""    
    return Flexibility_matrix(Domain, Roption=True)

class Diag_matrix (Structural_Matrix):
    """Block diagonal matrix of element flexibility/stiffness matrices for structural model
    

    this class represents the block diagonal matrix of element flexibility or stiffness matrices
    for a structural model.
    
    """
    tag = 'F'
    def __new__(cls, arry, rc_data, model):
        basic_forces = rc_data
        arry = np.asarray(arry).view(cls)
        arry.basic_forces = basic_forces
        arry.model = model
        # arry.column_data = [elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()]
        arry.column_data = arry.row_data = np.array([q.elem.tag+'_'+str(q.number) for q in basic_forces])
        # arry.row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()]
        return arry

    def __init__(self, arry, rc_data, model):
        """Parameters
        =========================================================================================
        model
        
        """
        self.model = model 
        pass

    def __matmul__(self, Vector):
        if type(Vector) is Deformation_vector:
            vect = np.matmul(self, Vector).view(iForce_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif isinstance(Vector, iForce_vector):
            vect = np.matmul(self, Vector).view(Deformation_vector)
            # vect.part = 'continuous'
            vect.row_data = self.row_data
            vect.matrix = self

        else:
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect

    @property
    def c(self):
        """Removes columns corresponding to element hinges/releases"""
        idx_c = self.model.idx_c
        newA = self[:,idx_c]
        newA = newA[idx_c,:]
        transfer_vars(self, newA)
        newA.basic_forces = self.basic_forces[idx_c]
        newA.column_data = self.column_data[idx_c]
        newA.row_data = self.row_data[idx_c]
        return newA

def Fs_matrix(model, Roption=True):
    """Returns a Flexibility_matrix object"""  
    if Roption:
        f  = np.array([elem.f_matrix(Roption) for elem in model.elems])
        basic_forces = np.array([q for elem in model.elems for q in elem.basic_forces if not q.rel]) 
    else:
        f  = np.array([elem.f_matrix(Roption) for elem in model.elems])
        basic_forces = np.array([q for elem in model.elems for q in elem.basic_forces]) 
    Fs = scipy.linalg.block_diag(*f) 
    
    basic_forces = basic_forces
    Fs = Diag_matrix(Fs, basic_forces, model)
    return Fs

def Ks_matrix(model):
    """Returns a Flexibility_matrix object"""  
    k  = np.array([elem.k_matrix() for elem in model.elems])
    Ks = scipy.linalg.block_diag(*k)  
    basic_forces = np.array([q for elem in model.elems for q in elem.basic_forces])
    Ks = Diag_matrix(Ks, basic_forces, model)
    return Ks


class Stiffness_matrix (Structural_Matrix):
    """...
    Parameters
    ============
    model
    
    -----------------------------------------------------------------------------------------
    """
    tag = 'K'
    def __new__(cls, arry, model, Roption=None):

        input_array = np.asarray(arry).view(cls)
        input_array.model = model
        input_array.column_data = [str(dof) for dof in range(1, model.nt+1)]
        input_array.row_data = ['P_{'+str(dof)+'}' for dof in range(1, model.nt+1)]
        
        return input_array

    def __init__(self, arry, model, Roption=None):
        self.subs = [None]
        pass
    

    def __matmul__(self, Vector):
        if type(Vector) is Displacement_vector:
            vect = np.matmul(self, Vector).view(nForce_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif type(Vector) is nForce_vector:
            vect = np.matmul(self, Vector).view(Displacement_vector)
            vect.row_data = self.row_data
            vect.model = self.model
        else:
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect

    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate([str(dof) for dof in range(1, self.model.nt+1)]) if int(dof) > self.model.nf]
        
        newK = np.delete(self, delrows, axis=0)
        newK = np.delete(newK, delrows, axis=1).view(type(self))
        transfer_vars(self, newK)
        
        newK.row_data = list(np.delete(self.row_data, delrows))
        newK.column_data = list(np.delete(self.column_data, delrows))
        return newK

def K_matrix(Model):
    """Returns a Stiffness_matrix object"""
    K = np.zeros((Model.nt, Model.nt))
    for elem in Model.elems:
        ke = elem.ke_matrix()
        for i, dof in enumerate(elem.dofs):
            for j, doff in enumerate(elem.dofs):
                K[int(dof)-1, int(doff)-1] += ke[i, j]

    return Stiffness_matrix(K, Model, Roption=None)

def K_tensor(Model,U=None):
    """Returns a Stiffness_matrix object"""
    K = np.zeros((Model.nt, Model.nt))
    for elem in Model.elems:
        ke = elem.ke_matrix(U)
        for i, dof in enumerate(elem.dofs):
            for j, doff in enumerate(elem.dofs):
                K[int(dof)-1, int(doff)-1] += ke[i, j]
    return K

def AssemblyTensor(Model):
    pass

def Kt_matrix(Model, State):
    """Returns a Stiffness_matrix object"""
    K = np.zeros((Model.nt, Model.nt))
    for elem in Model.elems:
        kt = elem.kt_matrix(State)
        for i, dof in enumerate(elem.dofs):
            for j, doff in enumerate(elem.dofs):
                K[int(dof)-1, int(doff)-1] += kt[i, j]

    return Stiffness_matrix(K, Model, Roption=None)


class Mass_matrix (Structural_Matrix):
    tag = 'M'
    def __new__(cls, Model):

        M = np.zeros((Model.nt, Model.nt))
        ddof = Model.ddof
        mass_dofs = []
        for node in Model.nodes:
            for i,dof in enumerate(node.dofs):
                if 'r' not in list(ddof.keys())[i]:
                    if node.mass != 0.0: mass_dofs.append(str(dof))
                    M[int(dof)-1, int(dof)-1] += node.mass


        input_array = np.asarray(M).view(cls)
        input_array.model = Model
        input_array.mass_dofs = mass_dofs
        input_array.column_data = ['u_{{'+str(dof)+'}}' for dof in range(1, Model.nt+1)]
        input_array.row_data = ['u_{{'+str(dof)+'}}' for dof in range(1, Model.nt+1)]
        return input_array

    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate([str(dof) for dof in range(1, self.model.nt+1)]) if int(dof) > self.model.nf]
        
        newK = np.delete(self, delrows, axis=0)
        newK = np.delete(newK, delrows, axis=1).view(type(self))
        transfer_vars(self, newK)
        
        newK.row_data = list(np.delete(self.row_data, delrows))
        newK.column_data = list(np.delete(self.column_data, delrows))
        return newK
    
    @property
    def m(self):
        delrows = [idx for idx, dof in enumerate([str(dof) for dof in range(1, self.model.nt+1)]) if int(dof) > self.model.nf]
        
        newK = np.delete(self, delrows, axis=0)
        newK = np.delete(newK, delrows, axis=1).view(type(self))
        transfer_vars(self, newK)
        
        newK.row_data = list(np.delete(self.row_data, delrows))
        newK.column_data = list(np.delete(self.column_data, delrows))
        return newK


class Displacement_vector(column_vector):
    tag = 'U'
    def __new__(cls, Kinematic_matrix, Vector=None):
        U = np.zeros((len(Kinematic_matrix.column_data)))
        return np.asarray(U).view(cls)

    def __init__(self, Kinematic_matrix, Vector=None):
        self.matrix = Kinematic_matrix 
        self.row_data = Kinematic_matrix.column_data
        self.subs = [None]
        if Vector is not None:
            for key in Vector.row_data:
                if key in self.row_data:
                    self.set_item(key, Vector.rows([key]))

    @property
    def f(self):
        """Removes rows corresponding to fixed dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.model.nf]
        newU = np.delete(self, delrows, axis=0).view(Displacement_vector)
        newU.row_data = list(np.delete(self.row_data, delrows))
        # newU.matrix = self.matrix
        newU.model = self.model
        return newU

class nDisplacement_vector(Structural_Vector):
    tag = 'U'
    def __new__(cls, arry, model, row_data, Vector=None):
        input_array = np.asarray(arry).view(cls)
        return input_array

    def __init__(self, arry, model, row_data, Vector=None):
        self.subs = [None]
        self.model = model
        self.row_data = row_data

        if Vector is not None:
            for key in Vector.row_data:
                if key in self.row_data:
                    self.set_item(key, Vector.rows([key]))

    @property
    def f(self):
        """Removes rows corresponding to fixed dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.model.nf]
        newU = np.delete(self, delrows, axis=0).view(nDisplacement_vector)
        newU.row_data = list(np.delete(self.row_data, delrows))
        newU.model = self.model
        return newU

def U_vector(model, vector=None):
    """Returns a Displacement_vector object"""
    U = np.zeros(model.nt)
    row_data = [str(dof) for dof in range(1,model.nt+1)]
    U = nDisplacement_vector(U, model, row_data)

    if vector is not None:
        if len(vector)==len(U):
            U[:,0] = vector[:]
        elif isinstance(vector,Structural_Vector):
            for key in vector.row_data:
                if key in U.row_data:
                    U.set_item(key, vector.rows([key]))
        else:
            U[:len(vector)] = vector.flatten()
    return U


class iForce_vector(Structural_Vector):
    tag = 'Q'
    def __new__(cls, arry, model, row_data, Vector=None):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, row_data, Vector=None):
        self.model = model
        self.subs = [None]
        self.row_data = row_data
        if Vector is not None:
            for key in Vector.row_data:
                if key in self.row_data:
                    self.set_item(key, Vector.get(key))
  #<
    @property
    def i(self):
        """Removes rows corresponding to redundant forces"""
        rdts = self.model.redundants

        tags = [q.elem.tag + str(q.nature) for q in rdts]
        delrows = [self.row_data.index(tag) for tag in tags]
        newQ = np.delete(self, delrows, axis=0).view(iForce_Vector)
        transfer_vars(self, newQ)
        newQ.subs.append('primary')
        newQ.row_data = list(np.delete(self.row_data, delrows))
        return newQ
    
    @property
    def c(self):
        """Remove rows corresponding to element hinges/releases"""
        idx_c = self.model.idx_c
        newQ = self[idx_c]
        transfer_vars(self, newQ)
        newQ.row_data = self.row_data[idx_c]
        return newQ
    
    @property
    def x(self):
        """Remove rows of corresponding to primary forces"""
        rdts = self.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        rows = [self.row_data.index(tag) for tag in tags]
        newV = self[rows]
        newV.row_data = [self.row_data[row] for row in rows]
        transfer_vars(self, newV)
        return newV

def Q_vector(model, vector=None):
    """Returns a iForce_vector object"""   
    
    arry = np.zeros((model.nQ,1))
    row_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
    Q = iForce_vector(arry, model, row_data)
    if vector is not None:
        if len(vector)==len(Q):
            Q[:,0] = vector[:]
        else:
            for key in vector.row_data:
                if key in Q.row_data:
                    Q.set_item(key, vector.rows([key]))
    return Q

def Q0_vector(model):
    """Returns a vector of initial element forces"""   
    arry = np.concatenate([elem.q0_vector() for elem in model.elems])
    row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.q.keys()] 
    return iForce_vector(arry, model, row_data)

def Qpl_vector(model):
    """Returns a vector of element plastic capacities""" 

    Qp_pos = [elem.Qp['+'][key] for elem in model.elems for key in elem.Qp['+']]
    Qp_neg = [elem.Qp['-'][key] for elem in model.elems for key in elem.Qp['-']]
    row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.Qp['-']]

    # del_idx = np.where(~Bf.any(axis=0))[0]
    # Qp_pos = np.delete(Qp_pos, del_idx)
    # Qp_neg = np.delete(Qp_neg, del_idx)
    # row_data = np.delete(row_data, del_idx)
    column_data = ['Q_{pl}^+', 'Q_{pl}^-']

    Qpl = nKinematic_matrix(np.array([Qp_pos, Qp_neg]).T, model, (row_data, column_data))
    return Qpl

def Qp_vector(model):
    """Returns a vector of element plastic capacities""" 
    B = B_matrix(model)
    Bf = B.f
    Qp_pos = [elem.Qp['+'][key] for elem in model.elems for key in elem.Qp['+']]
    Qp_neg = [elem.Qp['-'][key] for elem in model.elems for key in elem.Qp['-']]
    row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.Qp['-']]

    del_idx = np.where(~Bf.any(axis=0))[0]
    Qp_pos = np.delete(Qp_pos, del_idx)
    Qp_neg = np.delete(Qp_neg, del_idx)
    row_data = np.delete(row_data, del_idx)
    column_data = ['Q_{pl}^+', 'Q_{pl}^-']

    Qpl = nStatic_matrix(np.array([Qp_pos, Qp_neg]).T, model, (row_data, column_data))
    return Qpl

class nForce_vector(Structural_Vector):
    tag = 'P'
    def __new__(cls, arry, model, row_data, Vector=None):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, row_data, Vector=None):
        self.model = model
        self.subs = [None]
        self.row_data = row_data
        if Vector is not None:
            for key in Vector.row_data:
                if key in self.row_data:
                    self.set_item(key, Vector.get(key))


    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.model.nf]
        
        newP = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self, newP)
        # newB.model = self.model
        newP.row_data = list(np.delete(self.row_data, delrows))
        # newB.column_data = self.column_data
        return newP

    @property
    def d(self):
        """Removes rows corresponding to free dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) <= self.model.nf]
        newP = np.delete(self, delrows, axis=0).view(type(self))
        newP.model = self.model
        newP.row_data = list(np.delete(self.row_data, delrows))
        newP.subs.append('reactions')
        return newP
        
  #<

def P_vector(model, vector=None):
    P = np.zeros(model.nt)
    for node in model.nodes:
        p = node.p_vector()
        for i, dof in enumerate(node.dofs):
            P[int(dof)-1] += p[i]
    row_data = [str(dof) for dof in range(1, model.nt+1)]
    return nForce_vector(P, model, row_data)



# def P0_vector(model):
#     """Returns a _ object"""   
#     arry = np.concatenate([elem.p0_vector() for elem in model.elems]) 
#     row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.v.keys()] 
#     return nForce_vector(arry, model, row_data)

def P0_vector(model):
    P = np.zeros(model.nt)
    for elem in model.elems:
        dofs = elem.dofs
        if hasattr(elem, 'q0_vector'):
            p0 = elem.bg_matrix()@elem.q0_vector()
        else:
            p0 = np.zeros(len(dofs))

        for i,df in enumerate(dofs):
            P[int(df)-1] +=  p0[i]

    row_data = [str(dof) for dof in range(1, model.nt+1)]
    return nForce_vector(P, model, row_data)

def Pw_vector(model):
    P = np.zeros(model.nt)
    for elem in model.elems:
        dofs = elem.dofs
        if len(dofs)==6:
            P[int(dofs[0])-1] +=  elem.w['y']*elem.L/2*elem.sn
            P[int(dofs[1])-1] += -elem.w['y']*elem.L/2*elem.cs
            P[int(dofs[3])-1] +=  elem.w['y']*elem.L/2*elem.sn
            P[int(dofs[4])-1] += -elem.w['y']*elem.L/2*elem.cs
        else:
            pw = elem.pw_vector()
            for i,df in enumerate(dofs):
                P[int(df)-1] +=  pw[i]


    row_data = [str(dof) for dof in range(1, model.nt+1)]
    return nForce_vector(P, model, row_data)


class Deformation_vector(Structural_Vector):
    tag = "V"
    def __new__(cls, arry, model, row_data, Vector=None):
        input_array = np.asarray(arry).view(cls)
        return input_array

    def __init__(self, arry, model, row_data, Vector=None):
        self.model = model
        self.row_data = row_data
    
    @property
    def c(self):
        """Removes rows corresponding to element hinges/releases"""
        idx_c = self.model.idx_c
        newQ = self[idx_c]
        transfer_vars(self, newQ)
        newQ.row_data = self.row_data[idx_c]
        return newQ
    
    @property
    def i(self):
        """Removes rows corresponding to redundant forces"""
        rdts = self.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        delrows = [self.row_data.index(tag) for tag in tags]
        newV = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self,newV)
        newV.row_data = list(np.delete(self.row_data, delrows))
        newV.subs.append('primary')
        return newV

    @property
    def x(self):
        """Removes rows of corresponding to primary forces
        
        """
        rdts = self.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        rows = [self.row_data.index(tag) for tag in tags]
        newV = self[rows]
        newV.row_data = [self.row_data[row] for row in rows]
        transfer_vars(self, newV)
        return newV

def V_vector(model, vector=None):
    """Returns a Deformation_vector object"""    
    arry = np.zeros((model.nQ,1))
    row_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
    V = Deformation_vector(arry, model, row_data)
    if vector is not None:
        if len(vector)==len(V):
            V[:,0] = vector[:]
        else:
            for key in vector.row_data:
                if key in V.row_data:
                    V.set_item(key, vector.rows([key]))
    return V

def V0_vector(model):
    """Returns a Deformation_vector object"""   
    arry = np.concatenate([elem.v0_vector() for elem in model.elems]) 
    row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.v.keys()] 
    return Deformation_vector(arry, model, row_data)

def Aub_matrix(model, alpha):
    """Return the interaction upperbound matrix"""  
    aub  = np.array([elem.aub_matrix(alpha) for elem in model.elems])
    A = scipy.linalg.block_diag(*aub)
    return A



__version__ = "0.0.7"

try:
    import anabel.aisc
    import anabel.sections
    import anabel.analysis
except:
    pass

from anabel.matrices import *
from anabel.assemble import *

from anabel.graphics import *
#from anabel.solvers import *

#import anabel.utilities
#import anabel.matlib

#__all__ = anabel.assemblers.__all__ + ["elements", "assemblers", "graphics"]

settings = {
    "DATAFRAME_LATEX": True,
}



# Claudio Perez
# University of California, Berkeley
"""
# Templating

Wrappers, decorators and utilities for constructing expression templates.
"""
# Standard library imports
import inspect
from functools import wraps, partial
from collections import namedtuple,UserDict
from typing import Union, Tuple, Callable

# Optional imports
try:
    import jax
    from jax import jit
    from jaxlib import xla_client
except:
    jax = None
    jit=lambda x: x
    xla_client = None

import anon.diff
import anon.atom as anp
# import anabel.ops as anp
# import elle.iterate
from anon.core import interface

class Dual:
    pass


def template(
    dim:      Union[int,Tuple[Tuple[int,int]]]=None,
    statevar: str   = "state",
    main:     str   = "main",
    jacx:     str   = "jacx",
    form:     str   = "x,y,s=s,p=p,**e->x,y,s",
    params:   str   = "params",
    # paramvar: str   = "params",
    dimvar:   str   = None,
    origin:   tuple = None,
    order:    int   = 0,
    **kwargs
    # inv:Callable  = elle.iterate.inv_no1,
) -> Callable:
    """Decorator that wraps a basic local map generator.

    Attributes
    ----------
    `origin`:
        A structure containing arguments which may act as an 'origin' for the target function.

    `shape`: tuple
        Description of the shape of the dual and primal spaces.

    Generated arguments
    -------------------
    `_expose_closure`: bool
    :   Expose closed-over local variables as an attribute.

    `_jit`: bool
    :   If `True`{.python}, JIT-compiles the target function

    Studies
    -------
    [Structural dynamics (`elle-0008`)](/stdy/elle-0008)

    """


    def _decorator(func: Callable[...,dict]) -> Callable[...,tuple]:

        args:     str   = ",".join(["x","y","state"]+(["params"] if params else []))
        funcname = func.__name__.replace("<","_").replace(">","_") #+ ".domain"
        Arguments = namedtuple(funcname, args)

        @wraps(func)
        def wrapper(
            *args,
            _expose_closure=False,
            _jit=False,
            _reform:Union[str,bool]=False,
            _curry=False,
            _debug=False,
            _form = None,
            _iter=False,
            **kwds
        ) -> Callable:

            loc = func(*args, **kwds)
            if isinstance(order,str):
                _order = loc[order]
            else:
                _order = order

            if origin is None:
                assert dim is not None

                if isinstance(dim,str):
                    _dim = loc[dim]
                else:
                    _dim = dim

                if dim == 0:
                    xshape = yshape = 1

                elif isinstance(_dim,int):
                    xshape,yshape = ((_dim, 1), (_dim,1))
                else:
                    xshape, yshape = _dim

                    if xshape == 1:
                        pass
                    elif isinstance(xshape,int):
                        xshape = (xshape, 1)

                    if yshape == 1:
                        pass
                    elif isinstance(yshape,int):
                        yshape = (yshape, 1)

                shape = (xshape, yshape)
                if dim == 0:
                    x0, y0 = 0.0, 0.0
                else:
                    x0, y0 = anp.zeros(xshape), anp.zeros(yshape)
                    # shape = ((_order, *xshape), yshape)

                # _params = loc[params] if params else {}
                interface = anon.core.new_interface(
                    funcname+"_params",
                    loc[params] if params in loc else {} # get_unspecified_params()
                )

                # _params = interface(**loc[params])
                # _params = interface()
                
                _params = {}

                _origin = Arguments(*(
                    [x0, y0,loc[statevar],_params]
                ))

            else:
                # origin has been passed to generator
                if isinstance(origin,str):
                    _origin = loc[origin]
                else:
                    _origin = origin

                if isinstance(_origin[0], (float,int)):
                    xshape=1
                else:
                    xshape=_origin[0].shape

                if isinstance(_origin[1], (float,int)):
                    yshape=1
                else:
                    yshape=_origin[1].shape

                shape = xshape, yshape
                # TODO: refactor
                try:
                    _params = _origin[3]
                except:
                    _params = {}

                _params = anon.core.new_interface(
                    funcname+"_params",_params
                )

            if isinstance(main,str):
                _main = loc[main]
            else:
                _main = main



            #-transformations------------------
            if _reform is True:
                """
                assume main is of form `f: x->y`; reform to
                `f: (x,y,state,params,**kwds) -> (x,y,state)`
                """
                original_main = _main
                _main = lambda x, y=None,state={},params={},**kwds: (x, original_main(x), {})

            if _form == "<jax>":
                original_main = _main
                def _main(state, x):
                    # print(state, x)
                    out = original_main(x, None, state)
                    return out[2], out[1]


            if _jit:
                _main = jit(_main)
            #----------------------------------
            if _debug:
                def about():
                    print(f"order:  {_order}")
                    print(f"shape:  {shape}")
                    print(f"origin: {_origin}")
                _main.about = about
            if _expose_closure:
                _main.closure = loc
            #----------------------------------

            if "tanx" in loc: _main.tanx = loc["tanx"]
            _main.origin = _origin
            _main.params = _params
            _main.uparams = get_unspecified_parameters(_main)
            _main.shape  = shape
            _main.order  = _order
            if isinstance(jacx,str):
                if jacx in loc:
                    _jacx = loc["jacx"]
                elif jacx == "fwd":
                    _jacx = anon.diff.jacfwd(_main,1,0)
                else:
                    _jacx = anon.diff.jacfwd(_main,1,0)
            else:
                _jacx = jacx

            _main.jacx = jax.jit(_jacx) if _jit else _jacx


            return _main
            # ~ __invert__
        wrapper.part  = lambda **kwds: partial(wrapper, **kwds)
        return wrapper
    return _decorator

def wrap(f, *args, **kwds):
    """Wrap a pre-defined function to act as a local dual map.
    """
    state = kwds["state"] if "state" in kwds else {}
    params = kwds["params"] if "params" in kwds else {}
    return generator(
        *args, **kwds
    )(
        lambda *_, **__: {
            "main": f,
            "state": state,
            "params": params,
        }
    )(**kwds)

# def reform(func:Callable, form:str, newform:str):
#     old_args, old_out = form.split("->")

#     old_out = old_out.replace(" ","").split(",")


#     new_args, new_out = newform.split("->")
#     new_out = new_out.replace(" ","").split(",")

#     indices = ",".join(old_out.index[new] for new in new_out)

#     return exec(f"def reformed({}): return func({})[{indices}]")

def serialize(f):
    """
    >Requires JAX
    """
    return xla_client.XlaComputation(
        jax.xla_computation(f)(
            *f.origin
        ).as_serialized_hlo_module_proto()
    ).as_serialized_hlo_module_proto()


def assemble(f):
    backend = xla_client.get_local_backend()
    F = backend.compile(serialize(f))
    @wraps(f)
    def compiled_f(*args, **kwds):
        return xla_client.execute_with_python_values(
            F,(args, kwds), backend=backend
        )[0]
    return compiled_f



def generator_no2():
    """
    def F():
        name = "my-element"
        def main(x, y, state, **kwds, _name=name):
            pass
    """
    pass


def get_unspecified_parameters(func,recurse=False):
    """
    created 2021-03-31
    """
    signature = inspect.signature(func)
    params = {
        k: v
        for k, v in signature.parameters.items()
        if (v.default is None) or isinstance(v.default,inspect.Parameter) #not inspect.Parameter.empty
    }
    if recurse and "params" in signature.parameters:
        #print(signature.parameters)
        params.update({"params": {
                k: v
                for k,v in signature.parameters["params"].default.items()
                if (v is None) or isinstance(v,inspect.Parameter)
            }
        })
    return params

generator = template




import subprocess
import tempfile

import pytest

def _exec_notebook(path):
    with tempfile.NamedTemporaryFile(suffix=".ipynb") as fout:
        args = ["jupyter", "nbconvert", "--to", "notebook", "--execute",
                "--ExecutePreprocessor.timeout=1000",
                "--output", fout.name, path]
        subprocess.check_call(args)

@pytest.mark.parametrize("nb",["elle-0020"])
def test(nb):
    _exec_notebook(f"tests/{nb}.ipynb")

